from importlib import metadata
version = metadata.version(__package__)
date = '2024-02-06 18:00:00 GMT'
yaml_descriptions = """
- version : 0.5.2
  date : 2024-05-13
  comment : Patch to add METIS calibration sources.
  changes :
  - Add scopesim_templates/metis/laser.py and pinhole_mask.py

- version : 0.5.1
  date : 2024-02-06
  comment : Patch to remove Python 3.8 and SystemDict.
  changes :
  - Drop support for Python 3.8 #77
  - Use astar_utils.NestedMapping instead of scopesim.system_dict.SystemDict #76
  - Fix scaling of spectrum #74
  - Add config file for auto-generated release notes #68
  - Comment out print statements, perhaps replace with logging later #72

- version : 0.5.0
  date : 2023-10-19
  comment : Fix off-by-one error.
  changes :
  - Fix off-by-one error.
  - Use pyproject.toml.
  - Various cleanups and fixes.

- version : 0.4.4
  date : 2023-03-08
  comment : Fix metadata bug in 2D galaxy.

- version : 0.4.3
  date : 2023-02-15
  comment : Add pinhole mask and minor bugfixes

- version : 0.4.2
  date : 2022-07-12
  comment : Updated DLC server to point to scopesim.univie.ac.at

- version : 0.4.1
  date : 2022-04-22
  comment : Updated package structure
"""


"""Test for flatlamp."""

from astropy.io.fits import ImageHDU
from synphot import SourceSpectrum

from scopesim_templates.micado import flatlamp
from scopesim_templates.rc import Source


class TestFlatlamp:
    """Test for flatlamp."""

    def test_flatlamp_returns_source_object(self):
        """Test for flatlamp."""
        lamp = flatlamp()
        assert isinstance(lamp, Source)
        assert isinstance(lamp.spectra[0], SourceSpectrum)
        assert isinstance(getattr(lamp.fields[0], "field", lamp.fields[0]), ImageHDU)



# -*- coding: utf-8 -*-
"""Contains aux functions for clusters.py."""

from pathlib import Path
from collections.abc import Iterable

import numpy as np
from astropy.io.ascii import read as read_ascii
# from astropy.modeling.functional_models import KingProjectedAnalytic1D
from scipy.interpolate import interp1d

import pyckles


DIRNAME = Path(__file__).parent
MAMAJEK = read_ascii(DIRNAME / "mamajek_alt.dat", delimiter="|")
F_MASS2MV = interp1d(MAMAJEK["Msun"], MAMAJEK["Mv"], 1)
F_MASS2IDX = interp1d(MAMAJEK["Msun"], range(len(MAMAJEK["Msun"])), 0)
PICKLES = pyckles.SpectralLibrary("pickles")

tbl = PICKLES.catalog[1].data
mask_evol = tbl["evolution"] == 5
mask_metal = tbl["metalicity"] == "normal"
PICKLES_MS_V = tbl["name"][mask_evol * mask_metal]


def mass2spt(mass):
    """
    Find the closest spectral type for a given stellar mass.

    Parameters
    ----------
    mass : float or iterable of floats
        Star mass in units of solar mass.

    Returns
    -------
    spt : str or list of str
        Spectral type or list of spectral types.

    Notes
    -----
    `mass` should not be passed as a quantity.

    """
    if isinstance(mass, Iterable):
        # recursive
        return [mass2spt(m) for m in mass]

    idx = F_MASS2IDX(mass).astype(int)
    return MAMAJEK["SpT"][idx]


def mass2absmag(mass):
    """
    Interpolate the absolute V magnitude (Mv) for a given stellar mass.

    Parameters
    ----------
    mass : float or iterable of floats
        Star mass in units of solar mass.

    Returns
    -------
    absmag : float or list of floats
        Absolute magnitude or magnitudes.

    Notes
    -----
    `mass` should not be passed as a quantity.

    """
    if isinstance(mass, Iterable):
        # recursive
        return [mass2absmag(m) for m in mass]

    # Round to match interpolation precision, float to not have array.
    return float(F_MASS2MV(mass).round(3))


def closest_pickles(spt):
    """
    Retrieve closest spectral type available in the loaded Pyckles catalog.

    Parameters
    ----------
    spt : str or sequence of str
        Spectral type or types.

    Returns
    -------
    closest_pickle : str or list of str
        Closest spectral type or types in the catalog.

    """
    if isinstance(spt, Iterable) and not isinstance(spt, str):
        # recursive
        return [closest_pickles(s) for s in spt]

    to_strip = "OBAFGKMIV "

    lum_types = PICKLES_MS_V[PICKLES_MS_V.startswith(spt[0])]
    assert lum_types.strip(to_strip).isdigit().all()
    # The join is necessary to correctly parse 'M25V' as 2.5 and not 25...
    num_only = np.char.join(".", lum_types.strip(to_strip))
    classes = np.array(num_only).astype(float)

    idx = np.abs(classes - float(spt.strip(to_strip))).argmin()
    closest_pickle = lum_types[idx]

    return closest_pickle


# Unused function, commented out to not annoy pylint...
# def king_distribution(n, r_core, r_tidal):
#     king = KingProjectedAnalytic1D(r_core=r_core, r_tidal=r_tidal)
#     y = king(np.linspace(0, r_tidal, 100))

#     return None


def gaussian_distribution(n, fwhm, seed=None):
    if isinstance(seed, int):
        np.random.seed(seed)
    x, y = np.random.normal(loc=0, scale=fwhm/2.35, size=(2, n))
    return x, y



import warnings

from . import calibration
from . import extragalactic
from . import misc
from . import stellar
from . import utils
from . import metis

# This warning is emitted when just doing "import scopesim_templates", which
# should be a normal non-warning thing to do.
# TODO: Find a way to emit this warning only when the functions below are
#       accessed directly.
# warnings.warn("In a future version top level function calls will be removed. "
#               "Always use this syntax: from module.submodule import function",
#               DeprecationWarning, stacklevel=2)
from .misc.misc import source_from_image, source_from_file, source_from_cube
from .stellar.stars import star, stars, star_grid, star_field
from .stellar.clusters import cluster
from .extragalactic.galaxies import galaxy, galaxy3d, elliptical
from .calibration.calibration import empty_sky
from .micado import darkness, flatlamp
from . import micado



"""
Initial mass function.

KL:
   Copied (18.06.2020, vers:134c5ff) almost directly from
   https://github.com/astropy/PyPopStar/blob/master/popstar/imf/imf.py

J. Lu

Original code was taken from libimf package written by Jan Pflamm-Altenburg
and has been modified only marginally. The libimf code was licensed under
a GNU General Public License.

When I use this code, I should cite Pflamm-Altenburg & Kroupa 2006

Unfortunately, the code was almost completely un-commented, so all
comments are mine. I have also added substantially to the code to
make more convenient and general purpose functions.
"""

import time
import logging

import numpy as np

# TODO: standardize logging
log = logging.getLogger("imf")


class IMF():
    """
    IMF base class.

    The mass sampling and multiplicity implementation is here.

    Notes
    -----
    Code author: J. Lu.

    Original code was taken from libimf package written by Jan Pflamm-Altenburg
    (`Pflamm-Altenburg & Kroupa 2006
     <https://ui.adsabs.harvard.edu/abs/2006MNRAS.373..295P/abstract>`_)
    and has been modified only marginally, though more convinient and general
    purpose functions have been added. The libimf code was licensed under
    a GNU General Public License.


    Parameters
    ----------
    mass_limits : 2 element array; optional
        Define the minimum and maximum stellar masses in the IMF, in
        solar masses. First element is taken as the min, second element
        the max (e.g. `mass_limits` = [min_mass, max_mass]).

    multiplicity : SPISEA Multiplicity object or None
        If None, no multiplicity is assumed. Otherwise, use
        multiplicity object to create multiple star systems.
    """

    def __init__(self, mass_limits=None, multiplicity=None):
        if mass_limits is None:
            mass_limits = np.array([0.1, 150])
        self._multi_props = multiplicity
        self._mass_limits = mass_limits

        if multiplicity is None:
            self.make_multiples = False
        else:
            self.make_multiples = True

    def generate_cluster(self, totalMass, seed=None):
        """
        Generate a cluster of stellar systems with the specified IMF.

        Randomly sample from an IMF with specified mass
        limits until the desired total mass is reached. The maximum
        stellar mass is not allowed to exceed the total cluster mass.
        The simulated total mass will not be exactly equivalent to the
        desired total mass; but we will take one star above or below
        (whichever brings us closer to the desired total) the desired
        total cluster mass break point.

        Primary stars are sampled from the IMF, companions are generated
        based on the multiplicity properties provided.

        Parameters
        ----------
        totalMass : float
            The total mass of the cluster (including companions) in solar
            masses.

        seed: int
            If set to non-None, all random sampling will be seeded with the
            specified seed, forcing identical output.
            Default None

        Returns
        -------
        masses : numpy float array
            List of primary star masses.

        isMultiple : numpy boolean array
            List of booleans with True for each primary star that is in a
            multiple system and False for each single star.

        companionMasses : numpy float array
            List of
        """
        if (self._mass_limits[-1] > totalMass):
            log.info("sample_imf: Setting maximum allowed mass to %d",
                     totalMass)
            self._mass_limits[-1] = totalMass

        # Estimate the mean number of stars expected.
        self.normalize(totalMass)
        mean_number = self.int_xi(self._mass_limits[0], self._mass_limits[-1])
        newStarCount = np.round(mean_number)
        if self._multi_props is None:
            newStarCount *= 1.1

        # Generate output arrays.
        masses = np.array([], dtype=float)
        isMultiple = np.array([], dtype=bool)
        compMasses = []
        systemMasses = np.array([], dtype=float)

        # Loop through and add stars to the cluster until we get to
        # the desired total cluster mass.
        totalMassTally = 0
        loopCnt = 0

        # Set the random seed, if desired
        if seed:
            np.random.seed(seed=seed)

        while totalMassTally < totalMass:
            # Generate a random number array.
            uniX = np.random.rand(int(newStarCount))

            # Convert into the IMF from the inverted CDF
            newMasses = self.dice_star_cl(uniX)

            # Testing for Nans produced in masses
            test = np.isnan(newMasses)
            if np.sum(test) > 0:
                raise ValueError("Nan detected in cluster mass")

            # Dealing with multiplicity
            if self._multi_props is not None:
                # compMasses = [[] for newMass in newMasses]
                compMasses = np.empty((len(newMasses),), dtype=np.object)
                compMasses.fill([])

                # Determine the multiplicity of every star
                MF = self._multi_props.multiplicity_fraction(newMasses)
                CSF = self._multi_props.companion_star_fraction(newMasses)

                newIsMultiple = np.random.rand(int(newStarCount)) < MF

                # Copy over the primary masses. Eventually add the companions.
                newSystemMasses = newMasses.copy()

                #------New code-------#
                new = True
                if new:
                    # Function to calculate multiple systems more efficiently
                    #t1 = time.time()
                    compMasses, newSystemMasses, newIsMultiple = self.calc_multi(newMasses, compMasses,
                                                                                 newSystemMasses, newIsMultiple,
                                                                                    CSF, MF)

                    newTotalMassTally = newSystemMasses.sum()
                    isMultiple = np.append(isMultiple, newIsMultiple)
                    systemMasses = np.append(systemMasses, newSystemMasses)
                    #t2 = time.time()
                    #print("Generate multiples: {0}".format(t2 - t1))
                #------Previous code------#
                old = False
                if old:
                    # Calculate number and masses of companions
                    t1 = time.time()
                    for ii in range(len(newMasses)):
                        if newIsMultiple[ii]:
                            # determine number of companions
                            n_comp = 1 + np.random.poisson((CSF[ii]/MF[ii]) - 1)

                            # Determine the mass ratios of the companions
                            q_values = self._multi_props.random_q(np.random.rand(n_comp))

                            # Determine the masses of the companions
                            m_comp = q_values * newMasses[ii]

                            # Add in seperation information

                            # Only keep companions that are more than the minimum mass
                            compMasses[ii] = m_comp[m_comp >= self._mass_limits[0]]
                            newSystemMasses[ii] += compMasses[ii].sum()

                            # Double check for the case when we drop all companions.
                            # This happens a lot near the minimum allowed mass.
                            if len(compMasses) == 0:
                                newIsMultiple[ii] = False
                #-----------------------------------#
                    newTotalMassTally = newSystemMasses.sum()
                    isMultiple = np.append(isMultiple, newIsMultiple)
                    systemMasses = np.append(systemMasses, newSystemMasses)
                    t2 = time.time()
                    # print(F"All loop: {t2 - t1}")
            else:
                newTotalMassTally = newMasses.sum()

            # Append to our primary masses array
            masses = np.append(masses, newMasses)

            if (loopCnt >= 0):
                log.info("sample_imf: Loop %d added %.2e Msun to previous "
                         "total of %.2e Msun", loopCnt, newTotalMassTally,
                         totalMassTally)

            totalMassTally += newTotalMassTally
            newStarCount = mean_number * 0.1  # increase by 20% each pass
            loopCnt += 1

        # Make a running sum of the system masses
        if self._multi_props:
            massCumSum = systemMasses.cumsum()
        else:
            massCumSum = masses.cumsum()

        # Find the index where we are closest to the desired
        # total mass.
        idx = np.abs(massCumSum - totalMass).argmin()

        masses = masses[:idx+1]

        if self._multi_props:
            systemMasses = systemMasses[:idx+1]
            isMultiple = isMultiple[:idx+1]
            compMasses = compMasses[:idx+1]
        else:
            isMultiple = np.zeros(len(masses), dtype=bool)
            systemMasses = masses

        return (masses, isMultiple, compMasses, systemMasses)

    def calc_multi(self, newMasses, compMasses, newSystemMasses, newIsMultiple,
                   CSF, MF):
        """
        Calculate multiples more efficiently.

        Helper function; We will use array operations as much as possible.
        """
        # Identify multiple systems, calculate number of companions for
        # each
        idx = np.where(newIsMultiple == True)[0]
        n_comp_arr = 1 + np.random.poisson((CSF[idx] / MF[idx]) - 1)
        primary = newMasses[idx]

        # We will deal with each number of multiple system independently. This
        # is so we can put in uniform arrays in _multi_props.random_q.
        num = np.unique(n_comp_arr)
        for ii in num:
            tmp = np.where(n_comp_arr == ii)[0]

            if ii == 1:
                # Single companion case
                q_values = self._multi_props.random_q(np.random.rand(len(tmp)))

                # Calculate mass of companion
                m_comp = q_values * primary[tmp]

                # Only keep companions that are more than the minimum mass.
                # Update compMasses, newSystemMasses, and newIsMultiple
                good = np.where(m_comp >= self._mass_limits[0])[0]
                for jj in good:
                    compMasses[idx[tmp[jj]]] = np.transpose([m_comp[jj]])
                    newSystemMasses[idx[tmp[jj]]] += compMasses[idx[tmp[jj]]]

                bad = np.where(m_comp < self._mass_limits[0])[0]
                newIsMultiple[idx[tmp[bad]]] = False
            else:
                # Multple companion case
                q_values = self._multi_props.random_q(np.random.rand(len(tmp),
                                                                     ii))

                # Calculate masses of companions
                m_comp = np.multiply(q_values, np.transpose([primary[tmp]]))

                # Update compMasses, newSystemMasses, and newIsMultiple
                for jj in range(len(tmp)):
                    m_comp_tmp = m_comp[jj]
                    compMasses[idx[tmp[jj]]] = m_comp_tmp[m_comp_tmp >= self._mass_limits[0]]
                    newSystemMasses[idx[tmp[jj]]] += compMasses[idx[tmp[jj]]].sum()

                    # Double check for the case when we drop all companions.
                    # This happens a lot near the minimum allowed mass.
                    if len(compMasses[idx[tmp[jj]]]) == 0:
                        newIsMultiple[idx[tmp[jj]]] = False

        return compMasses, newSystemMasses, newIsMultiple


class IMF_broken_powerlaw(IMF):
    r"""
    Initialize a multi-part power-law with N parts.

    Each part of the power-law is described by a probability density function:

        P(m) \propto m ** power[n]

    for mass_limits[n] < m <= mass_limits[n+1].

    Parameters
    ----------
    mass_limits : numpy array
        Array of length (N + 1) with lower and upper limits of
        the power-law segments.

    coefficients : numpy array
        Array of length N that contains the powers for each
        power-law segment.

    multiplicity : Multiplicity object or None
        If None, no multiplicity is assumed. Otherwise, use
        multiplicity object to create multiple star systems.
    """

    def __init__(self, mass_limits, powers, multiplicity=None):
        if len(mass_limits) != len(powers) + 1:
            msg = ("Incorrect specification of multi-part powerlaw.\n"
                   "    len(massLimts) != len(powers)+1\n"
                   f"    {len(mass_limits) = }\n"
                   f"    {len(powers) = }\n")
            raise ValueError(msg)

        self._mass_limits = np.atleast_1d(mass_limits)
        self._m_limits_low = mass_limits[0:-1]
        self._m_limits_high = mass_limits[1:]
        self._powers = powers
        self._multi_props = multiplicity

        if multiplicity is None:
            self.make_multiples = False
        else:
            self.make_multiples = True

        # Calculate the coeffs to make the function continuous
        nterms = len(self._powers)
        coeffs = np.ones(nterms, dtype=float)

        # First term is just 1.0
        # Subsequent terms are products of previous terms and then some.
        for i in range(1, nterms):
            y = self._m_limits_low[i] ** self._powers[i-1]
            z = self._m_limits_low[i] ** self._powers[i]

            coeffs[i] *= coeffs[i-1] * y / z

        self.nterms = nterms
        self.coeffs = coeffs
        self.k = 1

    def xi(self, m):
        """
        Probability density describing the IMF.

        Input:
        m - mass of a star

        Output:
        xi - probability of measuring that mass.
        """
        return_float = isinstance(m, float)

        m = np.atleast_1d(m)
        xi = np.zeros(len(m), dtype=float)

        for i in range(len(xi)):
            tmp = gamma_closed(m[i], self._m_limits_low, self._m_limits_high)
            tmp *= self.coeffs * m[i]**self._powers
            y = tmp.sum()
            z = delta(m[i] - self._m_limits_high).prod()
            xi[i] = self.k * z * y

        if return_float:
            return xi[0]
        return xi

    def xi_new(self, m):
        """
        Probability density describing the IMF.

        Input:
        m - mass of a star

        Output:
        xi - probability of measuring that mass.
        """
        return_float = isinstance(m, float)

        m = np.atleast_1d(m)

        # Temporary arrays
        y = np.zeros(len(m), dtype=float)
        z = np.ones(len(m), dtype=float)

        # Loop through the different segments of the power law.
        for i in range(self.nterms):  # For i = 0 --> n, where n is the number of intervals
            aux = m - self._m_limits_low[i]  # Should this be i - 1?

            # Only continue for those entries that are in later segments
            idx = np.where(aux >= 0)[0]

            # Maybe we are all done?
            if len(idx) == 0:
                break

            m_tmp = m[idx]

            y_i = gamma_closed(m_tmp, self._m_limits_low[i],
                               self._m_limits_high[i])
            y_i *= self.coeffs[i] * m_tmp**self._powers[i]

            # Save results into the y array
            y[idx] += y_i

            z *= delta(m - self._m_limits_high[i])

        xi = self.k * z * y

        if return_float:
            return xi[0]
        return xi

    def m_xi(self, m):
        """Mass-weighted probability m*xi."""
        return_float = isinstance(m, float)
        m = np.atleast_1d(m)
        mxi = np.zeros(len(m), dtype=float)

        # Temporary arrays
        y = np.zeros(len(m), dtype=float)
        z = np.ones(len(m), dtype=float)

        # Loop through the different segments of the power law.
        for i in range(self.nterms):  # For i = 0 --> n, where n is the number of intervals
            aux = m - self._m_limits_low[i]  # Should this be i - 1?

            # Only continue for those entries that are in later segments
            idx = np.where(aux >= 0)[0]

            # Maybe we are all done?
            if len(idx) == 0:
                break

            m_tmp = m[idx]

            y_i = gamma_closed(m_tmp, self._m_limits_low[i],
                               self._m_limits_high[i])
            y_i *= self.coeffs[i] * m_tmp**(self._powers[i] + 1)

            # Save results into the y array
            y[idx] += y_i

            z *= delta(m - self._m_limits_high[i])

        mxi = self.k * z * y

        if return_float:
            return mxi[0]
        return mxi

    def getProbabilityBetween(self, massLo, massHi):
        """
        Return the integrated probability between some low and high mass value.
        """
        return self.int_xi(massLo, massHi)

    def int_xi(self, massLo, massHi):
        """
        Return the integrated probability between some low and high mass value.
        """
        return self.prim_xi(massHi) - self.prim_xi(massLo)

    def getMassBetween(self, massLo, massHi):
        """
        Return the integrated mass between some low and high mass value.
        """
        return self.int_mxi(massLo, massHi)

    def int_mxi(self, massLo, massHi):
        """
        Return the integrated total mass between some low and high stellar
        mass value. Be sure to normalize the IMF instance beforehand.
        """
        return self.prim_mxi(massHi) - self.prim_mxi(massLo)

    def prim_xi(self, a):
        """Helper function."""
        return_float = isinstance(a, float)

        a = np.atleast_1d(a)
        val = np.zeros(len(a), dtype=float)

        for i in range(len(val)):
            t1 = theta_open(a[i] - self._m_limits_high) * self.coeffs
            t2 = prim_power(self._m_limits_high, self._powers)
            t3 = prim_power(self._m_limits_low, self._powers)
            y1 = (t1 * (t2 - t3)).sum()

            t1 = gamma_closed(a[i], self._m_limits_low, self._m_limits_high)
            t1 *= self.coeffs
            t2 = prim_power(a[i], self._powers)
            t3 = prim_power(self._m_limits_low, self._powers)
            y2 = (t1 * (t2 - t3)).sum()

            val[i] = self.k * (y1 + y2)

        if return_float:
            return val[0]
        return val

    def prim_mxi(self, a):
        """Helper function."""
        return_float = isinstance(a, float)

        a = np.atleast_1d(a)
        val = np.zeros(len(a), dtype=float)

        for i in range(len(val)):
            t1 = theta_open(a[i] - self._m_limits_high) * self.coeffs
            t2 = prim_power(self._m_limits_high, self._powers+1)
            t3 = prim_power(self._m_limits_low, self._powers+1)
            y1 = (t1 * (t2 - t3)).sum()

            t1 = gamma_closed(a[i], self._m_limits_low, self._m_limits_high)
            t1 *= self.coeffs
            t2 = prim_power(a[i], self._powers+1)
            t3 = prim_power(self._m_limits_low, self._powers+1)
            y2 = (t1 * (t2 - t3)).sum()

            val[i] = self.k * (y1 + y2)

        if return_float:
            return val[0]
        return val

    def normalize(self, Mcl, Mmin=None, Mmax=None):
        """
        Normalize the IMF to a total cluster mass within a specified
        minimum and maximum stellar mass range.
        """
        self.k = 1.0
        self.Mcl = Mcl

        if Mmax is None:
            Mmax = self._m_limits_high[-1]

        if Mmin is None:

            Mmin = self._m_limits_low[0]

        if Mmax > Mcl:
            Mmax = Mcl

        if Mmax > self._m_limits_high[-1]:
            Mmax = self._m_limits_high[-1]

        if Mmin < self._m_limits_low[0]:
            Mmin = self._m_limits_low[0]

        self.norm_Mmin = Mmin
        self.norm_Mmax = Mmax

        self.k = Mcl / self.int_mxi(self.norm_Mmin, self.norm_Mmax)
        self.lamda = self.int_xi_cl(self._m_limits_low[0], self._mass_limits)

    def norm_cl_wk04(self, Mcl, Mmax=None, Mmin=None):
        """Helper function."""
        self.k = 1.0
        self.Mcl = Mcl

        if Mmax is None:
            Mmax = self._m_limits_high[-1]

        if Mmin is None:
            Mmin = self._m_limits_low[0]

        if Mmax > Mcl:
            Mmax = Mcl

        if Mmax > self._m_limits_high[-1]:
            Mmax = self._m_limits_high[-1]

        if Mmin < self._m_limits_low[0]:
            Mmin = self._m_limits_low[0]

        a = Mmin
        c = Mmax
        b = (c + a) / 2.0
        while (((c/b)-(a/b)) > 0.00001):
            mb = self.int_mxi(Mmin, b) / self.int_xi(b, Mmax)
            if mb < Mcl:
                a = b
            else:
                c = b
            b = (c + a) / 2.0

        Mmax = b
        self.norm_Mmin = Mmin
        self.norm_Mmax = Mmax

        self.k = Mcl / self.int_mxi(Mmin, Mmax)
        self.lamda = self.int_xi_cl(self._m_limits_low[0], self._mass_limits)

    def xi_cl(self, m):
        """Helper function."""
        return theta_closed(self.norm_Mmax - m) * self.xi(m)

    def mxi_cl(self, m):
        """Helper function."""
        return theta_closed(self.norm_Mmax - m) * self.m_xi(m)

    def int_xi_cl(self, left, right):
        """Helper function."""
        t1 = self.prim_xi(right)
        t2 = theta_closed(right - self.norm_Mmax)
        t3 = self.int_xi(self.norm_Mmax, right)
        t4 = self.prim_xi(left)
        t5 = theta_closed(left - self.norm_Mmax)
        t6 = self.int_xi(self.norm_Mmax, left)

        return (t1 - t2*t3) - (t4 - t5*t6)

    def int_mxi_cl(self, left, right):
        t1 = self.prim_mxi(right)
        t2 = theta_closed(right - self.norm_Mmax)
        t3 = self.int_mxi(self.norm_Mmax, right)
        t4 = self.prim_mxi(left)
        t5 = theta_closed(left - self.norm_Mmax)
        t6 = self.int_mxi(self.norm_Mmax, left)

        return (t1 - t2*t3) - (t4 - t5*t6)

    def dice_star_cl(self, r):
        """
        Given a list of random numbers (r), return a list of masses
        selected from the IMF.
        """
        return_float = isinstance(r, float)
        r = np.atleast_1d(r)  # Make sure it is an array

        x = r * self.lamda[-1]
        y = np.zeros(len(r), dtype=float)
        z = np.ones(len(r), dtype=float)

        # Loop through the different parts of the power law.
        for i in range(self.nterms):  # For i = 0 --> n, where n is the number of intervals
            aux = x - self.lamda[i]  # Should this be i - 1?

            # Only continue for those entries that are in later segments
            idx = np.where(aux >= 0)[0]

            # Maybe we are all done?
            if len(idx) == 0:
                break

            x_tmp = x[idx]
            aux_tmp = aux[idx]

            t1 = aux_tmp / (self.coeffs[i] * self.k)
            t1 += prim_power(self._m_limits_low[i], self._powers[i])
            y_i = gamma_closed(x_tmp, self.lamda[i], self.lamda[i+1])
            y_i *= inv_prim_power(t1, self._powers[i])

            # Save results into the y array
            y[idx] += y_i

            z *= delta(x - self.lamda[i+1])  # new version
            # z *= delta(x - self.lamda[i])   # old version

        if return_float:
            return y[0] * z[0]
        return y * z


class IMFSalpeter1955(IMF_broken_powerlaw):
    """
    Define IMF from `Salpeter (1955)
    <https://ui.adsabs.harvard.edu/abs/1955ApJ...121..161S/abstract>`_.
    Mass range is 0.4 M_sun - 10 M_sun.
    """

    def __init__(self, multiplicity=None):

        mass_limits = np.array([0.40, 10.0])
        powers = np.array([-2.3])

        super().__init__(mass_limits, powers, multiplicity=multiplicity)


class Miller_Scalo_1979(IMF_broken_powerlaw):
    """
    Define IMF from `Miller & Scalo (1979)
    <https://ui.adsabs.harvard.edu/abs/1979ApJS...41..513M/abstract>`_.
    Mass range is 0.1 M_sun - inf M_sun.
    """

    def __init__(self, multiplicity=None):
        mass_limits = np.array([0.1, 1, 10, np.inf])
        powers = np.array([-1.4, -2.5, -3.3])

        super().__init__(mass_limits, powers, multiplicity=multiplicity)


class Kennicutt_1983(IMF_broken_powerlaw):
    """
    Define IMF from `Kennicutt (1983)
    <https://ui.adsabs.harvard.edu/abs/1983ApJ...272...54K/abstract>`_.
    Mass range is 0.1 M_sun - inf M_sun.
    """

    def __init__(self, multiplicity=None):
        mass_limits = np.array([0.1, 1, np.inf])
        powers = np.array([-1.4, -2.5])

        super().__init__(mass_limits, powers, multiplicity=multiplicity)


class Kroupa_2001(IMF_broken_powerlaw):
    """
    Define IMF from `Kroupa (2001)
    <https://ui.adsabs.harvard.edu/abs/2001MNRAS.322..231K/abstract>`_.
    Mass range is 0.01 M_sun - inf M_sun.
    """

    def __init__(self, multiplicity=None):
        mass_limits = np.array([0.01, 0.08, 0.5, 1, np.inf])
        powers = np.array([-0.3, -1.3, -2.3, -2.3])

        super().__init__(mass_limits, powers, multiplicity=multiplicity)


class Weidner_Kroupa_2004(IMF_broken_powerlaw):
    """
    Define IMF from `Weidner & Kroupa (2004)
    <https://ui.adsabs.harvard.edu/abs/2004MNRAS.348..187W/abstract>`_.
    Mass range is 0.01 M_sun - inf M_sun.
    """

    def __init__(self, multiplicity=None):
        mass_limits = np.array([0.01, 0.08, 0.5, 1, np.inf])
        powers = np.array([-0.3, -1.3, -2.3, -2.35])

        super().__init__(mass_limits, powers, multiplicity=multiplicity)


##################################################
#
# Generic functions -- see if we can move these up.
#
##################################################
def prim_power(m, power):
    """
    Take floats or arrays, but return arrays.

    returns: m**(power + 1) / (power + 1) and handles the case when power = -1
    """
    return_float = isinstance(m, float) and isinstance(power, float)

    m = np.atleast_1d(m)
    power = np.atleast_1d(power)

    if (len(m) == 1) and (len(power) > 1):
        m = np.repeat(m, len(power))
    if (len(power) == 1) and (len(m) > 1):
        power = np.repeat(power, len(m))

    z = 1.0 + power
    val = (m**z) / z

    val[power == -1] = np.log(m[power == -1])

    if return_float:
        return val[0]
    return val


def inv_prim_power(x, power):
    """
    returns ((1+power) * x)**(1.0 / (1 + power)) and handles the case
    when power == -1.
    """
    return_float = isinstance(x, float) and isinstance(power, float)

    x = np.atleast_1d(x)
    power = np.atleast_1d(power)

    if (len(x) == 1) and (len(power) > 1):
        x = np.repeat(x, len(power))
    if (len(power) == 1) and (len(x) > 1):
        power = np.repeat(power, len(x))

    if x.shape != power.shape:
        raise ValueError()
        # FIXME: What was this before:
        # pdb.set_trace()

    z = 1.0 + power
    val = (z * x)**(1.0 / z)

    # --------------BUG CHECK---------------------#
    # This line doesn"t make sense if x is an N-element array and
    # power is just a 1-element array, which it appears to be for
    # imf.generate_cluster
    val[power == -1] = np.exp(x[power == -1])
    # -----------------------------------------------#
    if return_float:
        return val[0]
    return val


def log_normal(m, mean_logm, sigma_logm):
    return_float = isinstance(m, float) and isinstance(mean_logm, float) and \
        isinstance(sigma_logm, float)

    m = np.atleast_1d(m)
    mean_logm = np.atleast_1d(mean_logm)
    sigma_logm = np.atleast_1d(sigma_logm)

    z = np.log10(m) - mean_logm
    val = np.exp(-z**2 / (2.0 * sigma_logm**2)) / m

    if return_float:
        return val[0]
    return val


def prim_log_normal(m, mean_logm, sigma_logm):
    return_float = isinstance(m, float) and isinstance(mean_logm, float) and \
        isinstance(sigma_logm, float)

    m = np.atleast_1d(m)
    mean_logm = np.atleast_1d(mean_logm)
    sigma_logm = np.atleast_1d(sigma_logm)

    mu = (np.log10(m) - mean_logm) / (1.4142135623731 * sigma_logm)
    val = 2.88586244942136 * sigma_logm * error(mu)

    if return_float:
        return val[0]
    return val


def inv_prim_log_normal(m, mean_logm, sigma_logm):
    return_float = isinstance(m, float) and isinstance(mean_logm, float) and \
        isinstance(sigma_logm, float)

    m = np.atleast_1d(m)
    mean_logm = np.atleast_1d(mean_logm)
    sigma_logm = np.atleast_1d(sigma_logm)

    mu = inv_error(0.346516861952484 * m / sigma_logm)
    val = 10.0**(1.4142135623731 * sigma_logm * mu + mean_logm)

    if return_float:
        return val[0]
    return val


def mlog_normal(m, mean_logm, sigma_logm):
    return_float = isinstance(m, float) and isinstance(mean_logm, float) and \
        isinstance(sigma_logm, float)

    m = np.atleast_1d(m)
    mean_logm = np.atleast_1d(mean_logm)
    sigma_logm = np.atleast_1d(sigma_logm)

    z = np.log10(m) - mean_logm
    val = np.exp(-z**2 / (2.0 * sigma_logm**2))

    if return_float:
        return val[0]
    return val


def prim_mlog_normal(m, mean_logm, sigma_logm):
    return_float = isinstance(m, float) and isinstance(mean_logm, float) and \
        isinstance(sigma_logm, float)

    m = np.atleast_1d(m)
    mean_logm = np.atleast_1d(mean_logm)
    sigma_logm = np.atleast_1d(sigma_logm)

    eta = np.log10(m) - mean_logm - (sigma_logm**2 * 2.30258509299405)
    eta /= 1.4142135623731 * sigma_logm

    t1 = (1.15129254649702 * sigma_logm**2) + mean_logm

    val = error(eta)
    val *= 2.88586244942136 * sigma_logm * np.exp(2.30258509299405 * t1)

    if return_float:
        return val[0]
    return val


def theta_closed(x):
    """
    Pass in float or array of floats (x) and return 0 for x<0
    and 1 for everything else.
    """
    is_float = isinstance(x, float)

    x = np.atleast_1d(x)
    val = (x >= 0).astype("float")

    if is_float:
        return val[0]
    return val


def theta_open(x):
    """
    Pass in float or array of floats (x) and return 1 for x>0
    and 0 for everything else.
    """
    is_float = isinstance(x, float)

    x = np.atleast_1d(x)
    val = (x > 0).astype("float")

    if is_float:
        return val[0]
    return val


def delta(x):
    """
    Pass in float or array of floats (x) and return 0.5 for x==0
    and 1.0 for everything else.
    """
    is_float = isinstance(x, float)

    x = np.atleast_1d(x)
    val = np.ones(len(x), dtype=float)
    val[x == 0] = 0.5

    if is_float:
        return val[0]
    return val


def gamma_closed(m, left, right):
    """

    """
    return theta_closed(m - left) * theta_closed(right - m)


def error(x):
    x2 = x**2
    ax2 = 0.140012288686666 * x2

    val = np.sqrt(1.0 - np.exp(-x2*(1.27323954473516+ax2)/(1+ax2)))

    if x >= 0:
        return val
    return -val


def inv_error(x):
    x2 = x**2
    lnx2 = np.log(1.0 - x2)
    aux = 4.54688497944829 + (lnx2 / 2.0)
    y = -aux + np.sqrt(aux**2 - (lnx2 / 0.140012288686666))

    # val = np.sqrt(y)

    if x >= 0:
        return y
    return -y



from astar_utils import NestedMapping

from scopesim_templates import rc


class TestRC:
    def test_source_object_is_imported(self):
        assert hasattr(rc, "Source")

    def test_default_yaml_is_read_in(self):
        isinstance(rc.__config__, NestedMapping)



import pytest

import numpy as np
from astropy import units as u
from astropy.table import Table
from astropy.io.fits import ImageHDU
from synphot import SourceSpectrum

from scopesim_templates.stellar import star
from scopesim_templates.extragalactic import galaxy, galaxy3d, spiral_two_component
from scopesim_templates.misc.misc import source_from_array

# Add all initialied examples of sources to be tested to this list
SOURCE_LIST = [star(filter_name="Ks", amplitude=10*u.mag),
               star(filter_name="Paranal/HAWKI.J", amplitude=10*u.Jansky),
               spiral_two_component(),
               galaxy(sed="kc96/s0"),
               galaxy3d(sed="kc96/s0", ngrid=10),
               source_from_array(arr=np.ones(shape=(100, 100)),
                                 sed="kc96/s0",
                                 amplitude=15,
                                 pixel_scale=0.2,
                                 filter_curve="g")
               ]


# Run all tests on each source indivdually
@pytest.mark.parametrize("src", SOURCE_LIST)
class TestContentOfFields:
    """Test the <Source>.fields attribute to check that the content is correct."""

    def test_is_fields_a_list(self, src):
        assert isinstance(src.fields, list)

    def test_fields_list_has_only_imagehdus_or_tables(self, src):
        for field in src.fields:
            if not isinstance(field, Table):
                assert isinstance(getattr(field, "field", field), (Table, ImageHDU))

    def test_any_tables_have_correct_column_names(self, src):
        for field in src.fields:
            if isinstance(field, Table):
                fld = field
            else:
                if isinstance(getattr(field, "field", field), Table):
                    fld = getattr(field, "field", field)
                else:
                    continue

            req_colnames = ["x", "y", "ref", "weight"]
            assert all(col in fld.colnames for col in req_colnames)

    def test_any_imagehdus_have_correct_header_keywords(self, src):
        for field in src.fields:
            if isinstance(getattr(field, "field", field), ImageHDU):
                req_keys = ["SPEC_REF", "NAXIS", "NAXIS1", "NAXIS2",
                            "CUNIT1", "CUNIT2", "CTYPE1", "CTYPE2",
                            "CDELT1", "CDELT2", "CRVAL1", "CRVAL2",
                            "CRPIX1", "CRPIX2"]
                assert all(key in field.header for key in req_keys)


@pytest.mark.parametrize("src", SOURCE_LIST)
class TestContentOfSpectra:
    """Test the <Source>.spectra attribute to check that all spectra are useful."""

    def test_is_spectra_a_list_or_dict(self, src):
        # ScopeSim pre-0.9 is a list, afterwards dict. Allow both here.
        assert isinstance(src.spectra, (list, dict))

    def test_spectra_list_has_only_synphot_sourcespectrum_objects(self, src):
        if isinstance(src.spectra, list):
            for spectrum in src.spectra:
                assert isinstance(spectrum, SourceSpectrum)
        elif isinstance(src.spectra, dict):
            for spectrum in src.spectra.values():
                assert isinstance(spectrum, SourceSpectrum)
        else:
            raise TypeError(src.spectra)


@pytest.mark.parametrize("src", SOURCE_LIST)
class TestConnectionBetweenFieldsAndSpectra:
    """Test that all spectra referenced by .fields are in .spectra."""

    def test_all_spectra_in_table_ref_column_exist(self, src):
        for field in src.fields:
            if isinstance(getattr(field, "field", field), Table):
                for ref in field["ref"]:
                    assert isinstance(src.spectra[ref], SourceSpectrum)

    def test_all_spectra_refereced_in_imagehdu_header_exist(self, src):
        for field in src.fields:
            if isinstance(getattr(field, "field", field), ImageHDU):
                ref = field.header["SPEC_REF"]
                assert isinstance(src.spectra[ref], SourceSpectrum)



"""A standard cluster for MICADO."""

from ..utils.general_utils import add_function_call_str
from ..stellar import clusters


@add_function_call_str
def cluster():
    """A basic cluster."""
    return clusters.cluster(mass=10000, distance=2000, core_radius=.1)



"""Test cluster."""

from astropy.table import Table
from synphot import SourceSpectrum

from scopesim_templates.micado import cluster
from scopesim_templates.rc import Source


class TestCluster:
    """Test cluster."""

    def test_cluster_returns_source_object(self):
        """Test cluster."""
        sky = cluster()
        assert isinstance(sky, Source)
        assert isinstance(sky.spectra[0], SourceSpectrum)
        assert (isinstance(getattr(sky.fields[0], "field", sky.fields[0]), Table)
                or isinstance(sky.fields[0], Table))



from pytest import approx, mark
from collections.abc import Iterable
import numpy as np
from astropy.table import Table

import scopesim_templates.stellar.cluster_utils as cu

from matplotlib import pyplot as plt

PLOTS = False


class TestCatExists:
    def test_cat_is_table(self):
        assert isinstance(cu.MAMAJEK, Table)

    def test_cat_is_nonempty(self):
        assert len(cu.MAMAJEK) > 0


class TestMass2AbsMag:
    @mark.parametrize("mass, mv, atol",
                      [(2.3, 1.11, 0.1),
                       (200, -6, 0.2)])
    def test_returns_correct_value(self, mass, mv, atol):
        assert cu.mass2absmag(mass) == approx(mv, abs=atol)

    def test_returns_multiple_mvs_for_masses(self):
        mvs = cu.mass2absmag(np.array([200, 1.02]))
        assert isinstance(mvs, Iterable)
        assert mvs[1] == approx(4.79, rel=0.01)


class TestMass2SpT:
    @mark.parametrize("mass, spt",
                      [(250, "O5V"),
                       (2.3, "A0V"),
                       (1.02, "G2V"),
                       (0.01, "M9.5V")])
    def test_returns_correct_value(self, mass, spt):
        assert cu.mass2spt(mass) == spt

    def test_multiple_masses_at_once(self):
        spts = cu.mass2spt([2.3, 1.02])
        assert spts[1] == "G2V"
        assert isinstance(spts, list)


class TestClosestPickles:
    @mark.parametrize("spt_in, spt_out",
                      [("O1V", "O5V"),
                       ("G6V", "G5V"),
                       ("G7V", "G8V"),
                       ("M9V", "M6V"),
                       ("M2.4V", "M25V")])
    def test_individual_sources(self, spt_in, spt_out):
        assert cu.closest_pickles(spt_in) == spt_out

    @mark.parametrize("spts_in, spt_out",
                      [(["O1V", "M9V"], "O5V"),
                       (["O1V"]*10000, "O5V")])
    def test_multiple_sources(self, spts_in, spt_out):
        assert cu.closest_pickles(spts_in)[0] == spt_out


class TestGaussianDistribution:
    def test_returns_what_we_want(self):
        x, y = cu.gaussian_distribution(10000, 2.35)
        if PLOTS:
            plt.hist(x, bins=40)
            plt.show()

        # Apparently, 0.02 fails occasionally, so use 0.03.
        # TODO: Make these tests deterministic.
        assert np.std(x) == approx(1, rel=0.03)



"""Module should contain analytical extragalactic models."""

from dataclasses import dataclass
from collections.abc import Generator

import numpy as np

from astropy import units as u
from astropy.units import UnitsError
from astropy.modeling.core import Fittable2DModel
from astropy.modeling.parameters import Parameter
from astropy.modeling.models import Sersic2D


# Definitions

TWOPI = 2 * np.pi
FLOAT_EPSILON = float(np.finfo(np.float32).tiny)
GAUSSIAN_SIGMA_TO_FWHM = 2.0 * np.sqrt(2.0 * np.log(2.0))


class VelField(Fittable2DModel):
    r"""
    Two dimensional Velocity Field following arctan approximation.

    Parameters
    ----------
    ellip : float, u.Quantity
        Ellipticity on the sky
    theta : float, u.Quantity
        Position angle of the major axis. The rotation angle increases
        counterclockwise from the positive x axis.
    vmax : float, u.Quantity
        Constant rotation velocity for R>>rd.
    r_eff : float
        scale length of galaxy (assumed to be turnover radius).
    x0 : float, optional
        x position of the center.
    y0 : float, optional
        y position of the center.
    q : float, optional
        Disk thickness
    """

    vmax = Parameter(default=100)
    r_eff = Parameter(default=1)

    ellip = Parameter(default=0)    # maximum ellipticity 1 - q,  make tests
    theta = Parameter(default=0)

    x_0 = Parameter(default=0)
    y_0 = Parameter(default=0)

    q = Parameter(default=0.2)

    @classmethod
    def from_galaxy_base(cls, galaxy_base):
        """Create instance from ``GalaxyBase`` instance."""
        return cls(x_0=galaxy_base.x_0,
                   y_0=galaxy_base.y_0,
                   r_eff=galaxy_base.r_eff,
                   ellip=galaxy_base.ellip,
                   theta=galaxy_base.theta,
                   vmax=galaxy_base.vmax,
                   q=galaxy_base.q)

    @staticmethod
    def evaluate(x, y, vmax, r_eff,  ellip, theta, x_0, y_0, q):
        """Two dimensional velocity field, arctan approximation."""
        theta <<= u.deg
        theta = (-theta).to(u.rad)

        r_d = r_eff  # For now,  for n=1  r_eff = 1.678 * r_d
        # get inclination from ellipticity
        incl = np.arccos(np.sqrt(((1 - ellip) ** 2 - q ** 2) / (1 - q ** 2)))

        r = ((x - x_0) ** 2 + (y - y_0) ** 2) ** 0.5

        #   azimuthal angle in the plane of the galaxy = cos(theta) = cost
        cost = (-(x - x_0) * np.cos(theta) + (y - y_0) *
                np.sin(theta)) / (r + 0.00001)
        vrot = vmax*2 / np.pi*np.arctan(r/r_d)         # arctan model

        return vrot * np.sin(incl) * cost

    @property
    def input_units(self):
        if self.x_0.unit is None:
            return None
        return {"x": self.x_0.unit,
                "y": self.y_0.unit}

    def _parameter_units_for_data_units(self, inputs_unit, outputs_unit):
        # Note that here we need to make sure that x and y are in the same
        # units otherwise this can lead to issues since rotation is not well
        # defined.
        if inputs_unit["x"] != inputs_unit["y"]:
            raise UnitsError("Units of 'x' and 'y' inputs should match")
        return {"x_0": inputs_unit["x"],
                "y_0": inputs_unit["x"],
                "r_eff": inputs_unit["x"],
                "phi": u.deg,
                "vrot": outputs_unit["z"]}


class DispersionField(Fittable2DModel):
    r"""
    Two dimensional Velocity Dispersion.

    This follows a broken power law as in Veale et al. 2017

    .. math:: \sigma(R)=\sigma_0 2^{\gamma_1-\gamma_2} \left(\frac{R}{R_b}\right)^{\gamma_1}\left(1+\frac{R}{R_b}\right)^{\gamma_2-\gamma_1}

    where
    :math:`\sigma_0` is the velocity dispersion
    :math:`\gamma_1` and :math:`\gamma_2` are the inner and outer power slopes.
    Set to -0.04 and -0.42 as default values
    :math:`R_b` is the break radius, set to :math:`R_b = R_{eff}` for
    simplicity


    Parameters
    ----------
    incl : float, u.Quantity
        Inclination inclination between the normal to the galaxy plane and the
        line-of-sight.
    phi : float, u.Quantity
        Position angle of the major axis wrt to north (=up) measured
        counterclockwise.
    sigma : float, u.Quantity
        Velocity dispersion
    r_eff : float
        Scale length of galaxy (assumed to be turnover radius) it will be used
        as sigma.
    x0 : float, optional
        x position of the center.
    y0 : float, optional
        y position of the center.
    e_in : float, optional
        Inner power law slope.
    e_out : float, optional
        Outer power law slope.
    theta : float, u.Quantity
        Position angle of the major axis. The rotation angle increases
        counterclockwise from the positive x axis.
    """

    # incl = Parameter(default=45)
    ellip = Parameter(default=0)
    theta = Parameter(default=0)
    sigma = Parameter(default=100)
    r_eff = Parameter(default=1)
    x_0 = Parameter(default=0)
    y_0 = Parameter(default=0)
    e_in = Parameter(default=-0.0)
    e_out = Parameter(default=-0.5)

    @classmethod
    def from_galaxy_base(cls, galaxy_base):
        """Create instance from ``GalaxyBase`` instance."""
        return cls(x_0=galaxy_base.x_0,
                   y_0=galaxy_base.y_0,
                   r_eff=galaxy_base.r_eff,
                   ellip=galaxy_base.ellip,
                   theta=galaxy_base.theta,
                   sigma=galaxy_base.sigma,
                   e_in=galaxy_base.e_in,
                   e_out=galaxy_base.e_out)

    @staticmethod
    def evaluate(x, y, ellip, theta, sigma, r_eff, x_0, y_0, e_in, e_out):
        """Perform evaluation."""
        theta <<= u.deg
        theta = theta.to(u.rad)

        a, b = r_eff, (1 - ellip) * r_eff
        sin_theta, cos_theta = np.sin(theta), np.cos(theta)
        x_maj = (x - x_0) * cos_theta + (y - y_0) * sin_theta + 0.1
        x_min = -(x - x_0) * sin_theta + (y - y_0) * cos_theta + 0.1  # to avoid inf values
        z = np.sqrt((x_maj / a) ** 2 + (x_min / b) ** 2)
        result = sigma * 2**(e_in - e_out) * z**e_in * (1 + z)**(e_out - e_in)

        return result

    @property
    def input_units(self):
        if self.x_0.unit is None:
            return None
        return {"x": self.x_0.unit,
                "y": self.y_0.unit}

    def _parameter_units_for_data_units(self, inputs_unit, outputs_unit):
        # Note that here we need to make sure that x and y are in the same
        # units otherwise this can lead to issues since rotation is not well
        # defined.
        if inputs_unit["x"] != inputs_unit["y"]:
            raise UnitsError("Units of 'x' and 'y' inputs should match")
        return {"x_0": inputs_unit["x"],
                "y_0": inputs_unit["x"],
                "r_eff": inputs_unit["x"],
                "phi": u.deg,
                "sigma": outputs_unit["z"]}


@dataclass
class GalaxyBase:
    """
    Class to define a galaxy.

    It takes a set of parameters and creates the different moments for
    """

    x: np.ndarray
    y: np.ndarray
    x_0: int
    y_0: int
    amplitude: float
    r_eff: float
    ellip: float
    theta: float
    n: int = 4
    vmax: float = 0.0
    sigma: float = 0.0
    q: float = 0.2
    e_in: float = 0.04
    e_out: float = -0.42

    @property
    def intensity(self) -> np.ndarray:
        """
        2D light distribution following a Sersic2D profile.

        Returns
        -------
        result : numpy.ndarray
            Light distribution map.

        """
        mod = Sersic2D(x_0=self.x_0,
                       y_0=self.y_0,
                       amplitude=self.amplitude,
                       r_eff=self.r_eff,
                       n=self.n,
                       ellip=self.ellip,
                       theta=self.theta)
        return mod(self.x, self.y)

    @property
    def velocity(self) -> np.ndarray:
        """
        Velocity field according to the supplied parameters.

        Returns
        -------
        result : numpy.ndarray
            Velocity map.

        """
        if self.vmax <= 0.0:
            return np.ones_like(self.x)
        mod = VelField.from_galaxy_base(self)
        return mod(self.x, self.y)

    @property
    def dispersion(self) -> np.ndarray:
        """
        Velocity dispersion map according to the supplied parameters.

        Returns
        -------
        result : numpy.ndarray
            Velocity dispersion map.

        """
        if self.sigma <= 0.0:
            return np.ones_like(self.x)
        mod = DispersionField.from_galaxy_base(self)
        return mod(self.x, self.y)

    def regrid(self, ngrid: int = 10) -> np.ndarray:
        """
        Regrid the smooth velocity field.

        Regrid to regions with similar velocity and velocity dispersion.

        Parameters
        ----------
        ngrid : int, optional
            Griding factor. The default is 10.

        Returns
        -------
        sectors : numpy.ndarray
            A numpy array with sectors numbered.
        uniques : set
            Set of sector IDs.

        """
        velfield = self.velocity.value
        dispfield = self.dispersion.value

        vel_grid = np.round((ngrid // 2) * velfield / velfield.max())
        # Get order of magnitude for offset
        offset = 10**(int(np.log10(vel_grid.max())) + 2)
        sigma_grid = np.round((ngrid // 2 + 2) * dispfield /
                              dispfield.max()) * offset
        total_field = vel_grid + sigma_grid
        _, sectors = np.unique(total_field, return_inverse=True)
        uniques = set(sectors)
        # logger.debug("%d sectors", len(uniques))

        return sectors.reshape(total_field.shape), uniques

    def get_masks(self, ngrid: int = 10) -> Generator[np.ndarray]:
        """
        Return a generator of numpy masks from the regrided regions.

        Parameters
        ----------
        ngrid : TYPE, optional
            Griding factor, passed to ``GalaxyBase.regrid()``.
            The default is 10.

        Yields
        ------
        mask : numpy.ndarray
            Boolean array constructed from regrided regions.

        """
        grid, uniques = self.regrid(ngrid=ngrid)
        for value in uniques:
            yield np.equal(grid, value)



# -*- coding: utf-8 -*-
"""TBA."""
import pathlib

import numpy as np
import pyckles

from astropy import units as u
from astropy.io import fits
from astropy.utils import deprecated_renamed_argument
from astropy.wcs import WCS

from spextra import Spextrum

from ..rc import Source, im_plane_utils as ipu, \
    ter_curve_utils as tcu
from ..extragalactic import galaxy_utils as gal_utils
from ..extragalactic.exgal_models import GalaxyBase
from ..misc.misc import source_from_array
from ..utils import general_utils as gu
from ..utils.general_utils import add_function_call_str, hdu_to_synphot


__all__ = ["galaxy",
           "galaxy3d",
           "spiral_two_component",
           "elliptical"]


def _galaxy_setup(pixel_scale, r_eff, extend, **kwargs):
    """Construct GalaxyBase (wrapper), kwargs are passed to that class."""
    r_eff <<= u.arcsec
    r_eff_scaled = r_eff.value / pixel_scale.value

    image_size = 2 * r_eff_scaled * extend  # TODO: Needs unit check
    x_0 = image_size // 2
    y_0 = image_size // 2

    x, y = np.meshgrid(np.arange(image_size),
                       np.arange(image_size))

    gal = GalaxyBase(x=x, y=y, x_0=x_0, y_0=y_0, r_eff=r_eff_scaled,
                     amplitude=1, **kwargs)
    return gal


@deprecated_renamed_argument('plate_scale', 'pixel_scale', '0.1')
@add_function_call_str
def galaxy(sed,           # The SED of the galaxy
           z=0,             # redshift
           amplitude=15,           # magnitude
           filter_curve="g",        # passband
           pixel_scale=0.1,   # the plate scale "/pix
           r_eff=2.5,         # effective radius
           n=4,             # sersic index
           ellip=0.1,         # ellipticity
           theta=0,         # position angle
           extend=3,     # extend in units of r_eff
           ra=0.0,
           dec=0.0):
    """
    Create a source object of a galaxy described by its Sersic index and other parameters.

    This function is ideal for imaging or simple spectroscopy.

    Parameters
    ----------
    sed : str or Spextrum
    z : float
        redshift of the galaxy
    r_eff : float
        effective radius of the galaxy in arcsec, it accepts astropy.units
    amplitude : float
        magnitude or flux of the galaxy, it accepts astropy.units
    filter_curve : str
        name of the filter where the magnitude refer to
    pixel_scale : float
        the scale in arcsec/pixel of the instrument
    n : float
        Sersic index of the galaxy
    ellip : float
        ellipticity of the galaxy
    theta : float
        position angle of the galaxy
    extend : float
        Size of the image in units of r_eff
    ra : float, str
        RA of the source in degree, default 0.0
    dec : float, str
        DEC of the source in degree, default 0.0

    Returns
    -------
    src : scopesim.Source
    """
    pixel_scale <<= u.arcsec
    gal = _galaxy_setup(pixel_scale, r_eff, extend,
                        n=n, ellip=ellip, theta=theta)

    if isinstance(sed, str):
        spec = Spextrum(sed).redshift(z=z)
    elif isinstance(sed, Spextrum.__bases__):
        spec = sed.redshift(z=z)

    src = source_from_array(arr=gal.intensity, sed=spec,
                            pixel_scale=pixel_scale, amplitude=amplitude,
                            filter_curve=filter_curve, ra=ra, dec=dec)
    return src


def _get_masked_subsources(gal, ngrid, scaled_sp, header):
    masks = gal.get_masks(ngrid=ngrid)
    intensity = gal.intensity / gal.intensity.sum()
    velocity = gal.velocity
    dispersion = gal.dispersion

    for i, mask in enumerate(masks):
        # logger.debug("creating subsource %d", i)
        data = mask * intensity  # 0 ... 1

        med_vel = velocity[mask].mean()
        med_sig = dispersion[mask].mean()

        spec = scaled_sp.redshift(vel=med_vel).smooth(sigma=med_sig)

        header.update({"SPEC_REF": i})
        hdu = fits.ImageHDU(data=data, header=header)

        yield Source(image_hdu=hdu, spectra=spec)


@deprecated_renamed_argument('plate_scale', 'pixel_scale', '0.1')
@add_function_call_str
def galaxy3d(sed,           # The SED of the galaxy,
             z=0,             # redshift
             amplitude=15,           # magnitude
             filter_curve="g",        # passband
             pixel_scale=0.1,   # the plate scale "/pix
             r_eff=10,         # effective radius
             n=4,             # sersic index
             ellip=0.1,         # ellipticity
             theta=0,         # position angle
             vmax=100,
             sigma=100,
             extend=2,        # extend in units of r_eff
             ngrid=10):       # griding parameter
    """
    Create a simplified 3D map of a galaxy with flux, rotation velocity and velocity dispersion.

    The maps are binned according to the `ngrid` parameter, higher `ngrid`
    will create finer binned fields, but it may increase the computation time.

    The `ngrid` parameter does not specify the number of bins. A ngrid=10 will
    create around 40 independent regions whilst a `ngrid` of 100 will create
    around 2300 regions.

    This function is ideal for spectroscopy

    Parameters
    ----------
    sed : str or Spextrum
        SED of the galaxy, it can be a string or a Spextrum object.
    z : float
        Redshift of the galaxy.
    amplitude : float, u.Quantity
        Magnitude or flux of the galaxy. The spectrum will be re-escaled to
        this value.
    filter_curve : str
        Name of the filter where the `amplitude` is measured.
    pixel_scale : float
        The scale of the image in arcsec/pixel.
    r_eff : float
        Effective radius of the galaxy in arcsec. It accepts astropy.units.
    n : float
        Sersic index of the galaxy.
    ellip : float
        Ellipticity of the galaxy.
    theta : float
        Position angle of the galaxy.
    vmax : float
        Maximum rotation velocity of the galaxy.
    sigma : float
        Velocity dispersion of the galaxy.
    extend : float
        Size of the image in units of `r_eff`.
    ngrid : int
        Gridding parameter for creating of the galaxy.

    Returns
    -------
    src : scopesim.Source
    """
    vmax <<= u.km / u.s
    sigma <<= u.km / u.s
    amplitude <<= u.ABmag
    pixel_scale <<= u.arcsec

    gal = _galaxy_setup(pixel_scale, r_eff, extend,
                        n=n, ellip=ellip, theta=theta,
                        vmax=vmax, sigma=sigma)

    if isinstance(sed, str):
        shift_sp = Spextrum(sed).redshift(z=z)
        scaled_sp = shift_sp.scale_to_magnitude(amplitude=amplitude,
                                                filter_curve=filter_curve)
    elif isinstance(sed, Spextrum):
        scaled_sp = sed
    else:
        raise TypeError("Parameter 'sed' must be Spextrum, SourceSpectrum or "
                        f"str, found {type(sed)=}.")

    # TODO: don't think we really need to call intensity here
    w, h = gal.intensity.shape
    assert w == gal.x.shape[0], f"{w}, {gal.x.shape}"
    assert h == gal.y.shape[1], f"{h}, {gal.y.shape}"

    wcs_dict = {"NAXIS": 2,
                "NAXIS1": 2 * gal.x_0 + 1,
                "NAXIS2": 2 * gal.y_0 + 1,
                "CRPIX1": (w + 1) / 2,
                "CRPIX2": (h + 1) / 2,
                "CRVAL1": 0,
                "CRVAL2": 0,
                "CDELT1": -1 * pixel_scale.to(u.deg).value,
                "CDELT2": pixel_scale.to(u.deg).value,
                "CUNIT1": "deg",
                "CUNIT2": "deg",
                "CTYPE1": "RA---TAN",
                "CTYPE2": "DEC--TAN"}

    wcs = WCS(wcs_dict)

    header = fits.Header(wcs.to_header())
    header.update({"SPEC_REF": 0})

    return sum(_get_masked_subsources(gal, ngrid, scaled_sp, header),
               start=Source())


@add_function_call_str
def spiral_two_component(extent=60*u.arcsec, fluxes=(0, 0), offset=(0, 0)):
    """
    Create a spiral galaxy using NGC1232L as the template.

    Two components are included
        - the newer population (spiral arms), and
        - the older popultaion (bulge)

    Parameters
    ----------
    extent : float
        [arcsec]
    fluxes : list of floats, Quantity
        [mag | ABmag | jansky | FLAM | FNU | PHOTLAM | PHOTNU]
    offset : tuple of floats
        [arcsec]

    Returns
    -------
    gal : scopesim.Source
    """
    # params["object"] = "two component spiral galaxy"

    if isinstance(extent, u.Quantity):
        if extent.unit.physical_type == "angle":
            extent = extent.to(u.deg).value
        else:
            raise ValueError("Physical type of extent must be 'angle', but is:"
                             f" {extent.unit.physical_type}.")
    else:
        extent /= 3600.

    filename = "spiral_two_component.fits.bz2"
    path = pathlib.Path(__file__).parent.absolute() / filename
    with fits.open(path) as hdulist:
        img_ext = hdulist[0].header["IMG_EXT"]
        spec_ext = hdulist[0].header["SPEC_EXT"]

        srcs = []
        for i in range(img_ext, spec_ext):
            w, h = hdulist[i].data.shape
            hdulist[i].header["CRPIX1"] = (w + 1) / 2,
            hdulist[i].header["CRPIX2"] = (h + 1) / 2,
            hdulist[i].header["CRVAL1"] = 0
            hdulist[i].header["CRVAL2"] = 0
            hdulist[i].header["CDELT1"] = extent / w
            hdulist[i].header["CDELT2"] = extent / w
            hdulist[i].header["CUNIT1"] = "deg"
            hdulist[i].header["CUNIT2"] = "deg"
            hdulist[i].header["CTYPE1"] = "RA---TAN"
            hdulist[i].header["CTYPE2"] = "DEC--TAN"
            # hdulist[i].header["SPEC_REF"] = hdulist[i].header["SPEC_EXT"] - spec_ext

            i_spec_ext = spec_ext - img_ext + i
            srcs.append(Source(image_hdu=hdulist[i],
                        spectra=[hdu_to_synphot(hdulist[i_spec_ext])]))
    src = sum(srcs[1:], start=srcs[0])

    # TODO: scale image plane according to fluxes
    # TODO: shift header values according to offset

    # src.meta.update(params)
    # Ensure the number of _meta_dicts is the same as the number of _fields.
    # TODO: check if this still works with the new decorator...
    # src._meta_dicts += [{}] * (len(src.fields) - len(src._meta_dicts))

    return src


@deprecated_renamed_argument('magnitude', 'amplitude', '0.1')
@deprecated_renamed_argument('half_light_radius', 'r_eff', '0.1')
# @add_function_call_str
def elliptical(r_eff, pixel_scale, filter_name, amplitude,
               spectrum="NGC_0584", **kwargs):
    """
    Create a extended :class:`.Source` object for an elliptical galaxy.

    .. note:: This docstring is from simcado, needs to be updated

    Parameters
    ----------
    r_eff : float
        [arcsec]

    pixel_scale : float
        [arcsec]

    filter_name : str, TransmissionCurve, optional
        Default is "Ks". Values can be either:
        - the name of a SimCADO filter : see optics.get_filter_set()
        - or a TransmissionCurve containing a user-defined filter

    amplitude : float
        [mag, mag/arcsec2]

    spectrum : str, optional
        The spectrum to be associated with the galaxy. Values can either be:
        - the name of a SimCADO SED spectrum : see get_SED_names()
        - an EmissionCurve with a user defined spectrum


    Optional Parameters (passed to ``sersic_profile``)
    --------------------------------------------------
    n : float, optional
        Default = 4. Sersic index
        - n=1 for exponential (spiral),
        - n=4 for de Vaucouleurs (elliptical)

    ellipticity : float
        Default = 0.5

    angle : float
        [deg] Default = 30. Rotation anti-clockwise from the x-axis

    width, height : int
        [arcsec] Dimensions of the image. Default: 512*pixel_scale

    x_offset, y_offset : float
        [arcsec] The distance between the centre of the profile and the centre
        of the image. Default: (dx,dy) = (0,0)

    normalization : str, optional
        ["total", "half-light", "centre"] Where the profile equals unity
        If normalization equals:
        - "total" : [Default] whole image has brightness of ``amplitude`` [mag]
        - "half-light" : the pixels at the half-light radius have a surface
                         brightness of ``magnitude`` [mag/arcsec2]
        - "centre" : the maximum pixels have a surface brightness of
                     ``magnitude`` [mag/arcsec2]


    Returns
    -------
    galaxy_src : Source

    """
    # """
    # 1 make a sersic profile ImageHDU
    # 2 get spectrum from Brown
    # 3 scale spectrum to amplitude, assume ABmag if no unit
    # 4 make Source object
    # """

    # TODO: apply decorator here as well, somehow...
    params = {"n": 4,
              "ellipticity": 0.5,
              "angle": 30,
              "normalization": "total",
              "width": pixel_scale * 512,
              "height": pixel_scale * 512,
              "x_offset": 0,
              "y_offset": 0,
              "redshift": 0,
              "r_eff": r_eff,
              "pixel_scale": pixel_scale,
              "filter_name": filter_name,
              "amplitude": amplitude,
              "spectrum_name": str(spectrum),
              "rescale_spectrum": True}
    params.update(kwargs)
    params["function_call"] = gu.function_call_str(elliptical, kwargs=params)
    params["object"] = "elliptical galaxy"

    # 1 make a sersic profile ImageHDU
    img = gal_utils.sersic_profile(r_eff=r_eff / pixel_scale,    # everything in terms of pixels
                                   n=params["n"],
                                   ellipticity=params["ellipticity"],
                                   angle=params["angle"],
                                   normalization=params["normalization"],
                                   width=params["width"] / pixel_scale,
                                   height=params["height"] / pixel_scale)

    h_w, h_h = 0.5 * params["width"], 0.5 * params["height"]
    xs = (np.array([-h_w, h_w]) + params["x_offset"]) / 3600.
    ys = (np.array([-h_h, h_h]) + params["y_offset"]) / 3600.
    hdr = ipu.header_from_list_of_xy(xs, ys, pixel_scale / 3600.)
    hdu = fits.ImageHDU(data=img, header=hdr)

    # 2 get spectrum from Brown
    if isinstance(spectrum, str):
        brown_lib = pyckles.SpectralLibrary("brown", return_style="synphot")
        spectrum = brown_lib[spectrum]
        spectrum.z = params["redshift"]

    # 3 scale the spectra and get the weights
    if params["rescale_spectrum"]:
        if not isinstance(amplitude, u.Quantity):
            amplitude <<= u.ABmag
        spectrum = tcu.scale_spectrum(spectrum, filter_name, amplitude)

    # 4 make Source object
    src = Source(spectra=[spectrum], image_hdu=hdu)
    src.meta.update(params)

    return src



"""Test for flatlamp."""

import numpy as np
from astropy.table import Table
from astropy import units as u
from matplotlib import pyplot as plt
from synphot import SourceSpectrum

from scopesim_templates.micado.pinhole_masks import pinhole_mask
from scopesim_templates.rc import Source, load_example_optical_train

PLOTS = False


class TestPinholeMask:
    def test_pinhole_mask_returns_source_object(self):
        """ Example from pinhole_mask docstring"""
        dr = np.arange(-25, 26, 5)      # [arcsec]
        x, y = np.meshgrid(dr, dr)
        x, y = x.flatten(), y.flatten()
        waves = np.arange(0.7, 2.5, 0.001) * u.um

        ph_mask = pinhole_mask(x, y, waves, sum_factor=9001)

        assert isinstance(ph_mask, Source)
        assert isinstance(ph_mask.spectra[0], SourceSpectrum)
        assert (isinstance(getattr(ph_mask.fields[0], "field",
                                   ph_mask.fields[0]), Table)
                or isinstance(ph_mask.fields[0], Table))

    def test_pinhole_with_basic_instrument(self):
        dr = np.arange(-25, 26, 5)      # [arcsec]
        x, y = np.meshgrid(dr, dr)
        x, y = x.flatten(), y.flatten()
        waves = np.arange(0.7, 2.5, 0.001) * u.um

        src = pinhole_mask(x, y, waves, sum_factor=9001)
        opt = load_example_optical_train()
        opt.observe(src)

        im = opt.image_planes[0].data

        if PLOTS:
            plt.imshow(im)
            plt.show()

        assert np.max(im) > np.average(im)



from scopesim_templates import cluster
from scopesim_templates.rc import Source


def test_actually_works():
    my_cluster = cluster(mass=900, distance=50000, core_radius=1)
    assert isinstance(my_cluster, Source)



"""Currently only contains a simple zero-age open cluster."""

import numpy as np
import pyckles
from astropy import units as u
from astropy.table import Table
from astropy.utils import deprecated_renamed_argument

from spextra import Spextrum

from ..rc import __config__, Source
from ..utils.general_utils import function_call_str, RA0, DEC0
from . import imf, cluster_utils as cu


__all__ = ["cluster"]


@deprecated_renamed_argument("half_light_radius", "core_radius", "0.1")
def cluster(mass=1E3, distance=50000, core_radius=1, ra=RA0, dec=DEC0,
            **kwargs):
    """
    Generate a source object for a young cluster.

    The cluster distribution follows a gaussian profile with the
    ``core_radius`` corresponding to the HWHM of the distribution. The
    choice of stars follows a Kroupa IMF, with no evolved stars in the mix.
    Ergo this is more suitable for a young cluster than an evolved custer


    Parameters
    ----------
    mass : float
        [Msun] Mass of the cluster (not number of stars). Max = 1E5 Msun
    distance : float
        [pc] distance to the cluster
    core_radius : float
        [pc] half light radius of the cluster
    ra : float, str
        RA of the source
    dec : float, str
        DEC of the source

    Additional parameters
    ---------------------
    tidal_radius : float
        [pc] Not yet implemented, for later once there is a King profile
    multiplicity : Unknown
        A multiplicity object from the imf.py package. Not sure what it does.
    seed: float
        For setting the random seed for both masses and positions

    Returns
    -------
    src: scopesim.Source


    Examples
    --------
    Create a ``Source`` object for a young open cluster with half light radius
    of around 0.2 pc at the extragalactic centre and 1000 solar masses worth
    of stars:

        >>> from scopesim_templates.basic.stars import cluster
        >>> src = cluster(mass=1000, distance=8500, core_radius=0.2, seed=9001)

    """
    params = {"mass": mass,
              "distance": distance,
              "core_radius": core_radius,
              "tidal_radius": None,
              "multiplicity_object": None,
              "seed": __config__["!random.seed"],
              "ra": ra,
              "dec": dec}
    params.update(kwargs)
    params["function_call"] = function_call_str(cluster, kwargs=params)
    params["object"] = "cluster"

    # 1. sample masses from an IMF
    kroupa = imf.Kroupa_2001(params["multiplicity_object"])
    masses, _, _, _ = kroupa.generate_cluster(mass, seed=params["seed"])
    masses.clip(max=250, out=masses)

    # 2. get spec_types for masses
    spec_types = cu.mass2spt(masses)
    spec_types = cu.closest_pickles(spec_types)

    # 3. get spectra from pyckles
    pickles = pyckles.SpectralLibrary("pickles", return_style="synphot")
    unique_spts = set(spec_types)
    spectra = [pickles[spt] for spt in unique_spts]

    # 4. scale all spectra to V=0
    # spectra = [tcu.scale_spectrum(spec, "V", 0*u.mag) for spec in spectra]
    spectra = [Spextrum(modelclass=spec).scale_to_magnitude(filter_curve="V",
                                                            amplitude=0*u.mag)
               for spec in spectra]

    # 5. make ref list for spec_types from spectra
    ref = [np.where([spt == u_spt for u_spt in unique_spts])[0][0]
           for spt in spec_types]

    # 6. make weight list from Mv + dist_mod(distance)
    Mvs = np.array(cu.mass2absmag(masses))
    dist_mod = 5 * np.log10(distance) - 5
    weight = 10 ** (-0.4 * (Mvs + dist_mod))

    # 7. make x,y from half_light_radius and distance
    rad2arcsec = 206264.80624709636
    fwhm = 2 * core_radius / distance * rad2arcsec
    x, y = cu.gaussian_distribution(len(masses), fwhm=fwhm,
                                    seed=params["seed"])
    x = x << u.arcsec
    y = y << u.arcsec

    # 8. make table with (x,y,ref,weight)
    tbl = Table(names=["x", "y", "ref", "weight", "masses", "spec_types"],
                data=[x, y, ref, weight, masses, spec_types],
                units=[u.arcsec, u.arcsec, None, None, u.solMass, None])

    # 9. make Source with table, spectra
    src = Source(table=tbl, spectra=spectra)
    src.meta.update(params)

    return src



from astropy.table import Table
from synphot import SourceSpectrum

from scopesim_templates import calibration
from scopesim_templates.rc import Source


class TestEmptySky:
    def test_empty_sky_returns_source_object(self):
        sky = calibration.calibration.empty_sky()
        assert isinstance(sky, Source)
        assert isinstance(sky.spectra[0], SourceSpectrum)
        assert (isinstance(getattr(sky.fields[0], "field", sky.fields[0]), Table)
                or isinstance(sky.fields[0], Table))
        assert sky.fields[0]["ref"][0] == 0



from pathlib import Path

import numpy as np
from astropy.io import ascii
from astropy import units as u
import yaml

import pyckles

from ...stellar import stars
from ...extragalactic import elliptical
from ...rc import ter_curve_utils as tcu
from ...rc import Source


DATA_DIR = Path(__file__).parent


def viking_field(star_cat_id="illum", gal_cat_id="1", pixel_scale=0.004,
                 ra=None, dec=None, gal_field_size=60, gal_x=None, gal_y=None,
                 random_seed=None):
    """
    Make a field with background stars and galaxies according to Ric's input.

    Any combination of the star and galaxy catalogues is possible.


    Parameters
    ----------
    star_cat_id : str
        Default "illum". Options ["illum", "science", "stdstar"]

    gal_cat_id : str
        Default "1". Options ["1", "2", "3"]

    pixel_scale : float
        [arcsec] Default 0.004"

    ra, dec : float, None
        [deg] Default None. Coords for centre or star field.
        If None, field centre is taken as average of all star positions

    gal_field_size : float
        [arcsec] Default 60". Size of galaxy field for randomly assigning coords

    gal_x, gal_y : list of floats
        [arcsec] Default None.
        If None, galaxy coordinates are randomly assigned within a 1'x1' field

    random_seed : int, None
        For reproducible galaxy positions


    Returns
    -------
    viking_field : Source
        Combined source of all stars and galaxies


    Examples
    --------
    ::

        from scopesim_templates.calibration.micado import viking_fields as vf

        # Standard random star and galaxy generation at 0.004" resolution
        src = vf.viking_field(star_cat_id="illum", gal_cat_id="1")

        # For oversampling the galaxies and fixing their positions:
        src = vf.viking_field(star_cat_id="science", gal_cat_id="2",
                              pixel_scale=0.002, random_seed=9001)

    """
    star_src = load_stars_source(cat_id=star_cat_id, ra=ra, dec=dec)
    gal_src = load_galaxies_source(cat_id=gal_cat_id, pixel_scale=pixel_scale,
                                   box_size=gal_field_size, xs=gal_x, ys=gal_y)

    return gal_src + star_src


def load_stars_source(cat_id="illum", ra=None, dec=None):
    """
    Make a Source for one of Ric's Viking star fields.

    Parameters
    ----------
    cat_id : str
        Default "illum". Options ["illum", "science", "stdstar"]
    ra, dec : float, None
        [deg] Coords for centre or star field.
        If None, (RA, Dec) centre is taken from file header

    Returns
    -------
    src : Source

    """
    tbl = ascii.read(DATA_DIR / f"cat_{cat_id}.dat")
    tbl_meta = yaml.full_load("\n".join(tbl.meta["comments"]))

    if ra is None:
        ra0 = tbl_meta["RA_centre"]
    if dec is None:
        dec0 = tbl_meta["Dec_centre"]
    x, y = 3600 * (tbl["RA"] - ra0), 3600 * (tbl["Dec"] - dec0)

    mags = tbl["Hmag"].data
    src = stars(filter_name="H", amplitudes=mags, spec_types="A0V", x=x, y=y)

    return src


def load_galaxies_source(cat_id="1", pixel_scale=0.004, xs=None, ys=None,
                         angles=None, ellipticitys=None, ns=None, box_size=60,
                         random_seed=None):
    """
    Make a Source full of faint galaxies for one of Ric's background fields.

    .. warning: This funciton uses astropy's Sersic2D class, which can
       have unexpected consequences with GalFit (speak to Carmelo Archidiacono)


    Parameters
    ----------
    cat_id : str
        Default "1". Options ["1", "2", "3"]

    pixel_scale : float
        [arcsec] Default 0.004"

    xs, ys : list of floats
        [arcsec] : central positions of galaxies relative to field centre

    angles : list of floats
        [deg] rotation of galaxies

    ellipticitys : list of floats
        [0..1] 0=circle, 1=line

    ns : list of floats
        Sersic indicies

    box_size : float
        [arcsec] Default 60. Size of box for randomly assigned x,y positions

    random_seed : int, None
        For reproducible galaxy positions


    Returns
    -------
    gals_src : Source


    """
    if random_seed is not None:
        np.random.seed(random_seed)

    tbl = ascii.read(DATA_DIR / f"faintgal_{cat_id}.dat")

    h_mag = tbl["H_mag"].data.astype(float)             # [Vega mag]
    reff = tbl["Reff"].data.astype(float)               # [mas]
    gal_type = tbl["type"].data.astype(float)           # [0=ellip, 1=spiral]

    scale_factors = 10**(-0.4 * h_mag)
    r_effs = 1e-3 * reff                                # [arcsec]
    widths = 2 * r_effs

    # Or should there be a range > involved? E.g. Spiral 1-2, Ellip 3-5?
    # Ellip type 0, spiral type 1
    # Do spirals first
    if ns is None:
        n_gals = len(gal_type)
        ns = np.random.random(n_gals)   # Sersic index
        ns[gal_type == 1] += 1            # Spirals n=[1..2]
        ns[gal_type == 0] *= 2            # Ellips n=[3..5] => ([0..1] * 2) + 3
        ns[gal_type == 0] += 3

    if angles is None:
        angles = np.random.random(n_gals) * 360         # [deg]

    if ellipticitys is None:
        e_max = 0.85                                    # 0 = circular
        ellipticitys = e_max * np.random.random(n_gals)
        ellipticitys[gal_type == 0] = 0

    if xs is None:
        xs = np.random.random(n_gals) * box_size - 0.5 * box_size   # [arcsec]

    if ys is None:
        ys = np.random.random(n_gals) * box_size - 0.5 * box_size   # [arcsec]

    brown_lib = pyckles.SpectralLibrary("brown", return_style="synphot")
    spectrum = brown_lib["NGC_0584"]
    spectrum = tcu.scale_spectrum(spectrum, filter_name="H", amplitude=0*u.mag)

    srcs = []
    for i in range(n_gals):
        src = elliptical(r_eff=r_effs[i],
                         pixel_scale=pixel_scale,
                         filter_name="H",
                         amplitude=1,
                         spectrum=spectrum,
                         rescale_spectrum=False,
                         n=ns[i],
                         angle=angles[i],
                         ellipticity=ellipticitys[i],
                         width=widths[i],
                         height=widths[i],
                         normalization="total",
                         )
        src.fields[0].data *= scale_factors[0]
        src.shift(dx=xs[i], dy=ys[i])
        srcs += [src]

    gals_src = Source()
    gals_src.fields = [src.fields[0] for src in srcs]
    gals_src.spectra = [spectrum]
    gals_src._meta_dicts = [src.meta for src in srcs]

    return gals_src



"""Test for darkness."""

from astropy.table import Table
from synphot import SourceSpectrum

from scopesim_templates.micado import darkness
from scopesim_templates.rc import Source


class TestDarkness:
    """Test for darkness."""

    def test_darkness_returns_source_object(self):
        """Test for darkness."""
        sky = darkness()
        assert isinstance(sky, Source)
        assert isinstance(sky.spectra[0], SourceSpectrum)
        assert (isinstance(getattr(sky.fields[0], "field", sky.fields[0]), Table)
                or isinstance(sky.fields[0], Table))
        assert sky.fields[0]["ref"][0] == 0



import numpy as np
from astropy.modeling.models import Sersic2D


# TODO: could this be done with the existing implementation of GalaxyBase?
def sersic_profile(r_eff=100, n=4, ellipticity=0.5, angle=30,
                   normalization="total",
                   width=1024, height=1024, x_offset=0, y_offset=0,
                   oversample=1):
    """
    Return a 2D array with a normalised Sersic profile.

    Parameters
    ----------
    r_eff : float
        [pixel] Effective radius

    n : float
        Power law index.
        - n=1 for exponential (spiral),
        - n=4 for de Vaucouleurs (elliptical)

    ellipticity : float
        Ellipticity is defined as (a - b)/a. Default = 0.5

    angle : float
        [deg] Default = 30. Rotation anti-clockwise from the x-axis

    normalization : str, optional
        ["half-light", "centre", "total"] Where the profile equals unity
        If normalization equals:
        - "half-light" : the pixels at the half-light radius are set to 1
        - "centre" : the maximum values are set to 1
        - "total" : the image sums to 1

    width, height : int
        [pixel] Dimensions of the image

    x_offset, y_offset : float
        [pixel] The distance between the centre of the profile and the centre
        of the image

    oversample : int
        Factor of oversampling, default factor = 1. If > 1, the model is
        discretized by taking the average of an oversampled grid.


    Returns
    -------
    img : 2D array


    Notes
    -----
    Most units are in [pixel] in this function. This differs from
    :func:`.galaxy` where parameter units are in [arcsec] or [pc]

    """
    # Silently cast to integer
    os_factor = int(oversample)

    if os_factor <= 0:
        raise ValueError("Oversampling factor must be >=1.")

    width_os = os_factor * width
    height_os = os_factor * height
    x, y = np.meshgrid(np.arange(width_os), np.arange(height_os))

    dx = 0.5 * width_os + x_offset * os_factor
    dy = 0.5 * height_os + y_offset * os_factor

    r_eff_os = r_eff * os_factor

    mod = Sersic2D(amplitude=1, r_eff=r_eff_os, n=n, x_0=dx, y_0=dy,
                   ellip=ellipticity, theta=np.deg2rad(angle))
    img_os = mod(x, y)

    # Rebin os_factord image
    img = _rebin(img_os, os_factor)

    thresh = np.max([img[0, :].max(), img[-1, :].max(),
                     img[:, 0].max(), img[:, -1].max()])
    img[img < thresh] = 0

    if "cen" in normalization.lower():
        img /= np.max(img)
    elif "tot" in normalization.lower():
        img /= np.sum(img)

    return img


def _rebin(img, bpix):
    """Rebin image img by block averaging bpix x bpix pixels."""
    xedge = np.shape(img)[0] % bpix
    yedge = np.shape(img)[1] % bpix
    img_block = img[xedge:, yedge:]

    binim = np.reshape(img_block,
                       (int(img_block.shape[0]/bpix), bpix,
                        int(img_block.shape[1]/bpix), bpix))
    binim = np.mean(binim, axis=3)
    binim = np.mean(binim, axis=1)
    return binim



"""A collection of samplers for generating distributions."""

import numpy as np


def randomvariate(pdf, n, xmin, xmax):
    """
    Create random numbers according to an arbitrary PDF.

    Uses the rejection method for generating random numbers derived from an
    arbitrary probability distribution. For reference, see Bevington's book,
    page 84. Based on rejection*.py.
    Usage:
    >>> randomvariate(P,N,xmin,xmax)
    where
    P : probability distribution function from which you want to generate
    random numbers
    N : desired number of random values
    xmin,xmax : range of random numbers desired
    Returns:
    the sequence (ran,ntrials) where
    ran : array of shape N with the random variates that follow the input P
    ntrials : number of trials the code needed to achieve N
    Here is the algorithm:
    - generate x' in the desired range
    - generate y' between Pmin and Pmax (Pmax is the maximal value of your pdf)
    - if y'<P(x') accept x', otherwise reject
    - repeat until desired number is achieved
    Rodrigo Nemmen
    Nov. 2011
    TODO: Do optimization, e.g. using yield, etc
    """
    x = np.linspace(xmin, xmax, 10000)
    y = pdf(x)
    pmin = y.min()
    pmax = y.max()
    naccept = 0
    ntrial = 0
    ran = []
    while naccept < n:
        x = np.random.uniform(xmin, xmax)
        y = np.random.uniform(pmin, pmax)
        if y <= pdf(x):
            ran.append(x)
            naccept = naccept + 1
            #  print pdf(x)
            ntrial = ntrial + 1

    ran = np.asarray(ran)
    return ran, ntrial


def np_rvs(pdf, n, xmin, xmax, sampling=10000):
    """
    Numpy adaptation of randomvariate.

    This works in only one dimension
    This version is 100 to 1000 times faster than randomvariate for reasonable
    sized samples

    Parameters
    ----------
    pdf : TYPE
        Probability density function to sample.
    n : TYPE
        Number of returned random values.
    xmin, xmax : TYPE
        Range of random numbers desired.
    sampling : TYPE, optional
        Initial sampling of the parameter space. Increase for complex
        distributions. The default is 10000.

    Returns
    -------
    TYPE
        np.array with the values.

    """
    x = np.linspace(xmin, xmax, sampling)
    y = pdf(x)
    pmin = y.min()
    pmax = y.max()

    rel_area = (xmax - xmin) * (pmax - pmin) / np.trapz(y, x)
    m = int(n * rel_area) * 2
    ran = np.array([])

    while ran.size < n:
        X = np.random.uniform(xmin, xmax, m)
        Y = np.random.uniform(pmin, pmax, m)
        ran = np.append(ran, X[Y <= pdf(X)])  # the evaluation pdf(X) is the slow part for complex distributions

    return ran[:n]


def _metropolis_sampler(pdf, nsamples, xmin, xmax):
    """
    Taken from here:
    https://stackoverflow.com/questions/51050658/how-to-generate-random-numbers-with-predefined-probability-distribution

    Usage
        s1 = Sersic1D(amplitude=1, r_eff=5, n=4)
        p = lambda r: s1(r)
        samples = list(metropolis_sampler(p, N))
    """
    x = np.random.uniform(xmin, xmax)  # start somewhere

    for i in range(nsamples):
        trial = np.random.uniform(xmin, xmax)  # random neighbour from the proposal distribution
        acceptance = pdf(trial) / pdf(x)

        if np.random.uniform() < acceptance:
            x = trial

        yield x


def metropolis_sampler(pdf, nsamples, xmin, xmax):
    """
    This is a version of a metropolis sampler.

    Depending of the characteristics of the distribution it can be slower or
    comparable in speed to rvs
    """
    result = list(_metropolis_sampler(pdf=pdf, nsamples=nsamples,
                                      xmin=xmin, xmax=xmax))
    return np.array(result)



# -*- coding: utf-8 -*-
"""TBA."""

import logging
import functools
from inspect import signature

import numpy as np

from astropy import units as u
from astropy.wcs import WCS
from astropy.coordinates import SkyCoord

from synphot import SourceSpectrum, Empirical1D, ConstFlux1D

# TODO: Change this when scopesim is ready RA0, DEC0 = -10, 10
RA0, DEC0 = 0, 0


def hdu_to_synphot(hdu):

    wave = hdu.data["wavelength"]
    wave_unit = u.Unit(hdu.header["TUNIT1"])
    flux = hdu.data["flux"]
    flux_unit = u.Unit(hdu.header["TUNIT2"])

    assert len(wave) > 3, "At least 4 wavelength points are required."
    spec = SourceSpectrum(Empirical1D, points=wave*wave_unit,
                          lookup_table=flux*flux_unit)

    return spec


def vega_spectrum(mag=0):
    vega = SourceSpectrum.from_vega(cache=True)
    return vega * 10**(-0.4 * mag)


def st_spectrum(mag=0):
    return SourceSpectrum(ConstFlux1D, amplitude=mag*u.STmag)


def ab_spectrum(mag=0):
    return SourceSpectrum(ConstFlux1D, amplitude=mag*u.ABmag)


def _get_kwargs_and_defaults(args, kwargs, defaults):
    # BUG: This sometimes raises StopIteration, find out why and catch it!
    default_keys = iter(defaults)
    for arg in args:
        yield next(default_keys), arg
    for key in default_keys:
        yield key, kwargs.get(key, defaults[key].default)


def function_call_str(func, args=None, kwargs=None) -> str:
    """Generate function call signature string including default values."""
    args = args or ()
    kwargs = kwargs or {}
    defaults = signature(func).parameters
    call_kwargs = dict(_get_kwargs_and_defaults(args, kwargs, defaults))
    kwargs_repr = ", ".join(f"{k}={v!r}" for k, v in call_kwargs.items())
    func_str = f"{func.__module__}.{func.__qualname__}"
    call_str = f"{func_str}({kwargs_repr})"
    return call_str


def add_function_call_str(func):
    """Add constructor signature to meta of returned source. Decorator."""
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        call_str = function_call_str(func, args, kwargs)
        logging.debug("Calling %s ...", call_str)
        src = func(*args, **kwargs)
        src.meta["object"] = func.__name__
        src.meta["function_call"] = call_str
        return src
    return wrapper


def make_img_wcs_header(ra, dec, pixel_scale, image_size):
    """
    Create a WCS header for an image.

    ra : str, float
        "hh:mm:ss.s" or deg for the center of the image
    dec : str, float
        "dd:mm:ss.s or deg for the center of the image
    pixel_scale : float
        arcsecs
    image_size : tuple
        x, y where x, y are integers

    """
    if isinstance(ra, str):  # just assume is in the classical formats
        ra_unit = u.hourangle
    else:
        ra_unit = u.deg

    if isinstance(pixel_scale, u.Quantity):
        pixel_scale = pixel_scale.to(u.arcsec).value

    coords = SkyCoord(ra, dec, unit=(ra_unit, u.deg))
    x, y = image_size

    wcs = WCS(naxis=2)
    wcs.wcs.ctype = ["RA---TAN", "DEC--TAN"]
    wcs.wcs.cunit = [u.deg, u.deg]
    wcs.wcs.crpix = [(x + 1) / 2, (y + 1) / 2]
    wcs.wcs.cdelt = np.array([-pixel_scale / 3600, pixel_scale / 3600])
    wcs.wcs.crval = [coords.ra.value, coords.dec.value]

    wcs.wcs.cunit = [u.deg, u.deg]

    return wcs.to_header()


def make_cube_wcs_header():
    """TODO: Think if we need it."""
    raise NotImplementedError()



# -*- coding: utf-8 -*-
"""Contains simple source functions for testing and calibration."""

import warnings

import numpy as np
from astropy import units as u

from spextra import Spextrum

from ..rc import Source, __config__
from ..utils.general_utils import add_function_call_str
from ..misc.misc import uniform_source

__all__ = ["lamp",
           "flat_field",
           "empty_sky"]


@add_function_call_str
def lamp(waves, fwhm, fluxes) -> Source:
    """
    Create a homogenous source with a spectrum of emission lines.

    Emission line spectrum is superposed to a very faint continuum.

    Parameters
    ----------
    waves : np.array, u.Quantity
        a list or array of wavelengths
    fwhm : np.array, u.Quantity
       a list or array with the fwhm of the lines
    fluxes : np.array, u.Quantity
       a list or array with the fluxes of the lines

    Return
    ------
    src : Source

    Examples
    --------
    Create a ``Source`` lamp::

        >>> import numpy as np
        >>> import astropy.units as u
        >>> from scopesim_templates.calibration import lamp
        >>>
        >>> R = 20000   # Resolution of the instrument
        >>> waves = np.arange(1e4, 2.7e4, 100)
        >>> fwhm = 2.6*(waves/R)*np.ones(shape=waves.shape)  # Nyquist sampled
        >>> flux = 1e-10*np.ones(waves.shape)
        >>>
        >>> src = lamp(waves=waves, fwhm=fwhm, fluxes=flux)

    """
    # I don't understand this warning but it probably has a reason...
    warnings.warn("Make sure to turn off the appropriate ScopeSim effects "
                  "during the simulation", RuntimeWarning)

    waves <<= u.AA
    fwhm <<= u.AA

    w_minmax = [waves.min().value, waves.max().value] * u.AA
    # A very faint spextrum to add the lines
    spec = Spextrum.flat_spectrum(amplitude=40, waves=w_minmax)
    spec = spec.add_emi_lines(center=waves, fwhm=fwhm, flux=fluxes)
    src = uniform_source(sed=spec)
    # src.meta.update({"object": "lamp"})

    return src


@add_function_call_str
def flat_field(temperature=5000, amplitude=0*u.ABmag, filter_curve="V",
               extend=60) -> Source:
    """
    Create a flat-field source to be used in ``ScopeSim``.

    The spectral shape is given by a Black Body Spectrum with a temperature set
    by the user, so it can also be used in spectroscopy with enough realism.

    Flats-fields usually also contain an illumination pattern. They might be
    implemented in ScopeSim.effects.
    TODO: Investigate if that belongs eventually here.

    Default values are just a wild guess. We need to find more realistic ones.

    Parameters
    ----------
    temperature: float
        [Kelvin] Temperature of the lamp
    amplitude: u.Quantity
        [u.Quantity] amplitude of the lamp in u.ABmag, u.mag (vega) or u. STmag
    filter_curve: str
        any filter curve available for ``spextra``, used for scaling


    Returns
    -------
    src: Source

    """
    # I don't understand this warning but it probably has a reason...
    warnings.warn("Make sure to turn off the appropriate ScopeSim effects "
                  "during the simulation", RuntimeWarning)

    spec = Spextrum.black_body_spectrum(temperature=temperature,
                                        amplitude=amplitude,
                                        filter_curve=filter_curve)
    src = uniform_source(sed=spec, amplitude=amplitude,
                         filter_curve=filter_curve, extend=extend)
    # src.meta.update({"object": "flat_field"})

    return src


@add_function_call_str
def empty_sky() -> Source:
    """
    Return an empty source so that instrumental fluxes can be simulated.

    Returns
    -------
    sky : Source

    """
    sky = Source(lam=np.array([__config__["!spectral.wave_min"],
                               __config__["!spectral.wave_max"]]),
                 spectra=np.array([0, 0]), x=[0], y=[0], ref=[0], weight=[0])
    # src.meta.update({"object": "empty_sky"})

    return sky



from pathlib import Path
import sys
import yaml

from astar_utils import NestedMapping

# TODO: what's this?
if Path("F:/Work/ScopeSim/").exists():
    sys.path.append("F:/Work/ScopeSim/")

import scopesim

Source = scopesim.source.source.Source
ter_curve_utils = scopesim.effects.ter_curves_utils
im_plane_utils = scopesim.optics.image_plane_utils
scopesim_utils = scopesim.utils
load_example_optical_train = scopesim.load_example_optical_train

PKG_DIR = Path(__file__).parent
with (PKG_DIR / "defaults.yaml").open(encoding="utf-8") as file:
    __config__ = NestedMapping(yaml.full_load(file))



import numpy as np
from astropy.table import Table
from astropy import units as u
from astropy.utils.decorators import deprecated_renamed_argument

import pyckles
from spextra import Spextrum

from ..rc import __config__, Source
from ..utils.general_utils import RA0, DEC0, add_function_call_str,\
    function_call_str
from . import stars_utils as su


__all__ = ["star",
           "stars",
           "star_field",
           "star_grid",
           ]


def star_field(n, mmin, mmax, width, height=None, filter_name="V",
               ra=RA0, dec=DEC0, **kwargs):
    """
    Create a super basic field of stars with random positions and brightnesses.

    Parameters
    ----------
    n : int
        number of stars
    mmin, mmax : astropy.Quantity, float
        [mag, Jy] minimum and maximum flux amplitudes of the population
        in u.mag (u.ABmag) or u.Janksy where u is for astropy.units.
        If mmin and mmax are floats, Vega magnitudes are assumed
    width, height : float
        [arcsec] width of region to put stars in. if height=None, height=width
    filter_name : str
        For scaling the stars. Use either common names or Spanish-VO
        identifiers.
    ra : float or str
        RA of the center of the field (not used at the moment).
    dec : float or str
        DEC of the center of the field (not used at the moment).

    Additional parameters
    ---------------------
    x, y : lists, arrays
        [arcsec] The positions of the stars can be overridden by specifying the
        coordinates. The lists must contain N values

    Returns
    -------
    stars : scopesim.Source object
        A Source object with a field of stars that can be fed into the method:
        ``<OpticalTrain>.observe()``

    """
    params = {"n": n,
              "mmin": mmin,
              "mmax": mmax,
              "width": width,
              "height": height,
              "filter_name": "V",
              "seed": __config__["!random.seed"],
              "ra": ra,
              "dec": dec}
    params.update(kwargs)
    params["function_call"] = function_call_str(star_field, kwargs=params)
    params["object"] = "star field"

    if height is None:
        height = width

    if isinstance(params["seed"], int):
        np.random.seed(params["seed"])

    if not isinstance(mmin, u.Quantity) and not isinstance(mmax, u.Quantity):
        mmin, mmax = u.Quantity(mmin, u.mag), u.Quantity(mmax, u.mag)

    if "x" in kwargs and "y" in kwargs:
        x, y = kwargs["x"], kwargs["y"]
    else:
        rands = np.random.random(size=(2, n)) - 0.5
        x, y = width * rands[0], height * rands[1]

    # amplitudes = np.random.random(size=n) * (mmax - mmin) + mmin
    amplitudes = np.linspace(mmin, mmax, n)
    spec_types = ["A0V"] * n

    src = stars(filter_name=filter_name, amplitudes=amplitudes,
                spec_types=spec_types, x=x, y=y, ra=ra, dec=dec)
    src.meta["scaling_unit"] = mmin.unit
    src.meta.update(params)

    return src


@add_function_call_str
def star_grid(n, mmin, mmax, filter_name="V", separation=1, ra=RA0, dec=DEC0):
    """
    Create a square grid of A0V stars at equal magnitude intervals.

    Parameters
    ----------
    n : int
        the number of stars in the grid
    mmin : float
        [vega mag] the minimum (brightest) magntude of stars in the grid
    mmax : float
        [vega mag] maximum (faintest) magnitudes for stars in the grid
    filter_name : str
        any filter that is in the SimCADO package directory.
        See ``scopesim.optics.get_filter_set()``
    separation : float, optional
        [arcsec] an average speration between the stars in the grid can be
        specified. Default is 1 arcsec
    ra : float or str
        RA of the center of the field (not used at the moment)
    dec : float or str
        DEC of the center of the field (not used at the moment)

    Returns
    -------
    src : ``scopesim.Source``

    """
    side_len = int(np.ceil(np.sqrt(n)))
    x = separation * (np.arange(n) % side_len - (side_len - 1) / 2)
    y = separation * (np.arange(n) // side_len - (side_len - 1) / 2)

    src = star_field(n, mmin, mmax, side_len, filter_name=filter_name,
                     x=x, y=y, ra=ra, dec=dec)
    return src


@deprecated_renamed_argument("mags", "amplitudes", "0.1")
@add_function_call_str
def stars(filter_name, amplitudes, spec_types, x, y, library="pyckles",
          ra=RA0, dec=DEC0):
    """
    Create a scopesim.Source object for a list of stars with given amplitudes.

    .. note:: If amplitudes have no units, vega magnitudes are assumed

    Parameters
    ----------
    filter_name : str
        For scaling the stars. Use either common names or Spanish-VO
        identifiers.
    amplitudes : list of Quanitity, float
        [mag, Jy] amplitudes for the list of stars. Acceptable astropy.units:
        [u.mag, u.ABmag, u.Janksy]. If no units are given, Vega magnitudes are
        assumed
    spec_types : list of strings
        the spectral type(s) of the stars, e.g. "A0V", "G5III"
    x, y : arrays of float
        [arcsec] x and y coordinates of the stars on the focal plane
    ra, dec : float
        coordinates of the center of the field
    library: str
        Library where the spectroscopic types are taken. By default are taken
        from the pickles library using the `pyckles` package. Other libraries
        available are kurucz, bosz/lr, bosz/mr, bosz/hr, etc for MIR coverage
        and different spectral resolutions. Please see the `spextra` package
        for more information

    Returns
    -------
    src : ``scopesim.Source``

    Examples
    --------
    Create a ``Source`` object for a random group of stars::

        >>> import numpy as np
        >>> import astropy.units as u
        >>> from scopesim_templates.stellar import stars
        >>>
        >>> n = 100
        >>> spec_types = ["A0V", "G2V", "K0III", "M5III", "O8I"]
        >>> ids = np.random.randint(0, 5, size=n)
        >>> star_list = [spec_types[i] for i in ids]
        >>> mags = np.random.normal(loc=20, scale=3, size=n) * u.mag
        >>> x, y = np.random.random(size=(2, n))
        >>>
        >>> src = stars("Ks", mags, spec_types, x, y)

    .. note: All positions are in arcsec.

    The final positions table is kept in the ``<Source>.fields`` attribute::

        >>> src.fields[0]

    Each star in this table has an associated spectrum kept in the
    ``<Source>.spectra`` attribute. These stars are connected to the spectra in
    the list by the "ref" column in the ``.fields`` table::

        >>> src.spectra

    The stars can be scaled in units of u.mag, u.ABmag or u.Jansky. Any filter
    listed on the spanish VO filter profile service can be used for the scaling
    (``http://svo2.cab.inta-csic.es/theory/fps/``). SVO filter names need to be
    in the following format ``observatory/instrument.filter``::

        >>> amplitudes = np.linspace(1, 3631, n) * u.Jansky
        >>> filter_name = "Paranal/HAWKI.Ks"
        >>> stars(filter_name, amplitudes, spec_types, x=x, y=y)

    """
    if not isinstance(spec_types, (list, tuple, np.ndarray)):
        spec_types = [spec_types]

    if not isinstance(amplitudes, u.Quantity):
        amplitudes = u.Quantity(amplitudes, u.mag, copy=False)

    if not isinstance(x, u.Quantity):
        x = u.Quantity(x, u.arcsec, copy=False)
    if not isinstance(y, u.Quantity):
        y = u.Quantity(y, u.arcsec, copy=False)

    unique_types = list(set(spec_types))
    if library == "pyckles":
        pickles_lib = pyckles.SpectralLibrary("pickles",
                                              return_style="synphot")
        cat_spec_types = su.nearest_spec_type(unique_types, pickles_lib.table)
        # print(cat_spec_types)
    # scale the spectra and get the weights
    if amplitudes.unit in [u.mag, u.ABmag, u.STmag]:
        zero = 0 * amplitudes.unit
        weight = 10**(-0.4*amplitudes.value)
    else:
        zero = 1 * amplitudes.unit
        weight = amplitudes.value

    if library == "pyckles":
        # spectra = [tcu.scale_spectrum(pickles_lib[spt], filter_name, zero)
        #            for spt in zip(cat_spec_types)]
        spectra = [Spextrum(modelclass=pickles_lib[spt]).scale_to_magnitude(
            filter_curve=filter_name, amplitude=zero)
                   for spt in cat_spec_types]

    else:
        spectra = [Spextrum(library + "/" + spec.lower()).scale_to_magnitude(
            amp, filter_curve=filter_name)
                   for spec, amp in zip(spec_types, amplitudes)]
        weight = np.ones(shape=amplitudes.shape)

    # get the references to the unique stellar types
    ref = np.array([unique_types.index(i) for i in spec_types])

    if len(ref) == 1:
        ref = ref[0] * np.ones(len(x), dtype=int)
        spec_types *= len(x)

    tbl = Table(names=["x", "y", "ref", "weight", "spec_types"],
                data=[x, y, ref, weight, spec_types])

    src = Source(spectra=spectra, table=tbl)
    return src


def star(filter_name, amplitude, spec_type="A0V", x=0, y=0, library="pyckles"):
    if isinstance(amplitude, u.Quantity) is False:
        amplitude = amplitude * u.mag

    src = stars(filter_name, [amplitude.value] * amplitude.unit,
                [spec_type], [x], [y], library=library)

    return src


star.__doc__ = stars.__doc__



from pytest import approx
import numpy as np

import astropy.units as u
import synphot as sp

from scopesim_templates.stellar import star, stars, star_field, star_grid
from scopesim_templates.rc import Source, ter_curve_utils as tcu


def source_eq(source_lhs: Source, source_rhs: Source):
    """hacky way to ensure two source"""
    eq = len(source_lhs.spectra) == len(source_rhs.spectra)
    if isinstance(source_lhs.spectra, list):
        for spectrum_lhs, spectrum_rhs in zip(source_lhs.spectra,
                                              source_rhs.spectra):
            eq = eq and all(spectrum_lhs.waveset == spectrum_rhs.waveset)
    elif isinstance(source_lhs.spectra, dict):
        for spectrum_lhs, spectrum_rhs in zip(source_lhs.spectra.values(),
                                              source_rhs.spectra.values()):
            eq = eq and all(spectrum_lhs.waveset == spectrum_rhs.waveset)
    return eq


class TestStar:
    def test_star_stars_equivalent(self):
        """verify that star and stars yield the same object and have compatible interfaces"""
        src_from_star = star("Generic/Johnson.V", 0, "A0V", 0, 0)
        src_from_stars = stars("Generic/Johnson.V", [0], ["A0V"], [0], [0])
        src_from_star_unit = star("Generic/Johnson.V", 0*u.mag, "A0V", 0, 0)
        src_from_stars_unit = stars("Generic/Johnson.V", [0*u.mag], ["A0V"], [0], [0])

        _ = source_eq(src_from_star, src_from_stars)
        assert source_eq(src_from_star, src_from_stars)
        assert source_eq(src_from_star, src_from_star_unit)
        assert source_eq(src_from_star, src_from_stars_unit)


class TestStars:
    def test_returns_correct_photon_counts_when_scaled_to_vega_zero(self):
        src = stars("Generic/Johnson.V", [0]*u.mag, ["A0V"], [0], [0])

        filt = tcu.get_filter("Generic/Johnson.J")
        obs = sp.Observation(src.spectra[0], filt)
        assert obs.effstim(sp.units.PHOTLAM).value == approx(193, rel=0.01)

        filt = tcu.get_filter("Generic/Johnson.I")
        obs = sp.Observation(src.spectra[0], filt)
        assert obs.effstim(u.Jansky).value == approx(2367, rel=0.01)

    def test_returns_correct_photon_count_when_initialised_in_hawki_j(self):
        src = stars("Paranal/HAWKI.J", [0]*u.mag, ["A0V"], [0], [0])
        filt = tcu.get_filter("Paranal/HAWKI.J")
        obs = sp.Observation(src.spectra[0], filt)
        assert obs.effstim(sp.units.PHOTLAM).value == approx(210, rel=0.01)

    def test_returns_correct_photon_count_when_scaled_to_jansky(self):
        src = stars("Generic/Johnson.V", [3631]*u.Jansky, ["A0V"], [0], [0])
        filt = tcu.get_filter("Paranal/HAWKI.J")
        obs = sp.Observation(src.spectra[0], filt)
        phs = obs.effstim(sp.units.PHOTLAM).value * src.fields[0]["weight"][0]
        assert phs == approx(210, rel=0.01)

    def test_only_includes_unique_spectra(self):
        spts = ["A0V"]*3 + ["K1II"]*2
        src = stars("Paranal/HAWKI.J", ([0]*5)*u.mag, spts, [0]*5, [0]*5)
        assert len(src.spectra) == 2
        assert np.max(src.fields[0]["ref"]) == 1
        assert np.all(src.fields[0]["weight"] == 1)


class TestStarField:
    def test_returns_source_object_with_minimal_input(self):
        src = star_field(n=2, mmin=0, mmax=5, width=1)
        assert isinstance(src, Source)


class TestStarGrid:
    def test_returns_source_object_with_correct_weighting_without_unit(self):
        src = star_grid(n=3, mmin=0, mmax=5)
        assert isinstance(src, Source)
        assert src.fields[0]["weight"][2] == 0.01
        assert len(src.spectra) == 1

    def test_returns_source_object_with_correct_weighting_in_jansky(self):
        src = star_grid(n=3, mmin=1*u.Jansky, mmax=3631*u.Jansky)
        assert src.fields[0]["weight"][2] == 3631



# -*- coding: utf-8 -*-
"""Pinhole mask for METIS."""

from itertools import product

import numpy as np
from astropy import units as u

from ..utils.general_utils import add_function_call_str
from ..micado.pinhole_masks import pinhole_mask as micado_mask


@add_function_call_str
def pinhole_mask(
        nx=5,
        ny=28,
        dx=0.45,
        dy=0.27945,
        waves=np.arange(3, 6, .001)*u.um,
        temperature = 1500*u.K,
        **kwargs
):
    """
    Create pinhole mask source for METIS, simulated as a grid of point sources.

    This currently uses the (flexible)  MICADO pinhole mask internally.

    Parameters
    ----------
    nx : int, optional
        Number of points along x axis. The default is 5.
    ny : int, optional
        Number of points along y axis. The default is 28.
    dx : float, optional
        (Half) extension in x direction. The default is 0.45.
    dy : float, optional
        (Half) extension in y direction. The default is 0.27945.
    waves : array-like, optional
        Wavelength axis. The default is np.arange(3, 6, .001)*u.um.
    **kwargs
        Any other kwargs passed to micado.pinhole_mask.

    Returns
    -------
    TYPE
        DESCRIPTION.

    """
    x, y = zip(*product(np.linspace(-dx, dx, nx), np.linspace(-dy, dy, ny)))

    return micado_mask(x, y, waves, temperature, **kwargs)




from pathlib import Path

import numpy as np
from scipy import signal
from astropy.io import ascii as ioascii
from astropy.io import fits
from astropy import units as u
from synphot import SourceSpectrum, Empirical1D
from synphot.units import PHOTLAM

from ..rc import Source, im_plane_utils

DATA_DIR = Path(__file__).parent / "data"


def line_list(unit_flux=1*PHOTLAM,
              dwave=0.001,
              smoothing_fwhm=0.003,
              width=15.,
              height=1.,
              filename="masterlinelist_2.1.txt",
              ):
    """
    Make a ScopeSim.Source object for the calibration lamp line list.

    [PHOTLAM] = [ph s-1 cm-2 AA-1]

    Parameters
    ----------
    unit_flux : astropy.Quantity
        Default 1*PHOTLAM.
        Flux conversion factor for a relative intensity value of 1.
        The unit can be overridden with any equivalent astropy Quantity.
    dwave : float
        Bin width in units of input file
    smoothing_fwhm : int, float
        FWHM of smoothing kernel in units of ``dwave`` (i.e. input file)
    filename : str
        Name of line list file found in ``micado/data``

    Returns
    -------
    line_list_src : scopesim.Source

    Examples
    --------
    ::

        from scopesim_templates.micado import spectral_calibrations as mic_spec
        from synphot.units import PHOTLAM
        from astropy import units as u

        # Use all defaults
        src = mic_spec.line_list()

        # Make line list brighter
        src = mic_spec.line_list(unit_flux=2000)        # [PHOTLAM] by default
        src = mic_spec.line_list(unit_flux=5*PHOTLAM)
        src = mic_spec.line_list(unit_flux=300*u.Unit("ph s-1 m-2 um-1))

        # Smooth with a gaussian kernel where FWHM is 3 bins wide
        src = mic_spec.line_list(smoothing_fwhm=3)

    """
    if not isinstance(unit_flux, u.Quantity):
        unit_flux *= PHOTLAM

    # import line list and pad with zeros
    wave, flux = import_line_spectrum(filename, dwave)

    if smoothing_fwhm is not None and isinstance(smoothing_fwhm, (int, float)):
        sigma = smoothing_fwhm / dwave / 2.35
        kernel = signal.windows.gaussian(M=int(8 * sigma + 1), std=sigma)
        kernel /= kernel.sum()
        flux = signal.convolve(flux, kernel, mode="same")

    # make spectrum
    assert len(wave) > 3, "At least 4 wavelength points are required."
    spec = SourceSpectrum(Empirical1D, points=wave * u.nm,
                          lookup_table=flux * unit_flux)

    # make unity image that covers MICADO FOV of +/- 30" arcsec
    dw = 0.5 * width / 3600         # input needed in [deg]
    dh = 0.5 * height / 3600         # input needed in [deg]
    pixel_scale = 0.1 * dw

    hdr = im_plane_utils.header_from_list_of_xy([-dw, dw], [-dh, dh],
                                                pixel_scale)  # [deg]
    hdr["SPEC_REF"] = 0
    im = np.ones((hdr["NAXIS1"], hdr["NAXIS2"]))
    field = fits.ImageHDU(header=hdr, data=im)

    line_list_src = Source(image_hdu=field, spectra=[spec])

    return line_list_src


def import_line_spectrum(filename, dwave=0.0001):
    """
    Read in and pad the line list into a spectrum.

    Parameters
    ----------
    filename : str
        Name of line list file found in ``micado/data``

    Returns
    -------
    wave_filled : np.ndarray
        wavelength vector in units of whatever is in the input file
    flux_filled : np.ndarray
        flux vector in units of whatever is in the input file

    """
    line_tbl = ioascii.read(DATA_DIR / filename)

    flux = line_tbl["Relative_Intensity"].data
    wave = line_tbl["Wavelength"].data
    wave = np.round(wave / dwave) * dwave       # round to level of dwave
    w0, w1 = wave[0], wave[-1]
    wave_filled = np.arange(w0 - dwave, w1 + 2 * dwave, dwave)
    flux_filled = np.zeros_like(wave_filled)

    for w, f in zip(wave, flux):
        i = int((w - w0) / dwave)
        flux_filled[i] += f

    return wave_filled, flux_filled



# This test suite takes long because it creates a 10 million entry spectrum 5x over

from pytest import approx

import numpy as np
from astropy import units as u
from astropy.io import fits
import matplotlib.pyplot as plt

from synphot import SourceSpectrum
from scopesim_templates.micado import spectral_calibrations as mic_sc

PLOTS = False


class TestLineList:
    def test_returns_source_object_with_everything_as_needed(self):
        src = mic_sc.line_list()

        wave = np.arange(0.8, 2.5, 1e-6) * u.um
        flux = src.spectra[0](wave)

        if PLOTS:
            plt.plot(wave, flux)
            plt.show()

        assert isinstance(src.spectra[0], SourceSpectrum)
        assert isinstance(getattr(src.fields[0], "field", src.fields[0]), fits.ImageHDU)

    def test_returns_flux_scaled_spectrum(self):
        src1 = mic_sc.line_list()
        src2 = mic_sc.line_list(unit_flux=5)

        wave = np.arange(0.81, 2.5, 1e-7) * u.um
        flux1 = src1.spectra[0](wave).value
        flux2 = src2.spectra[0](wave).value

        assert 5 * flux1.sum() == approx(flux2.sum(), rel=1e-4)

    def test_returns_a_smoothed_spectrum(self):
        src1 = mic_sc.line_list(smoothing_fwhm=None)
        src2 = mic_sc.line_list(smoothing_fwhm=0.005)

        wave = np.arange(0.81, 2.5, 1e-7) * u.um
        flux1 = src1.spectra[0](wave).value
        flux2 = src2.spectra[0](wave).value

        assert flux1.max() > flux2.max()
        assert flux1.sum() == approx(flux2.sum(), rel=2e-2)

    # def test_micado_lines(self):
    #
    #     import numpy as np
    #     from astropy import units as u
    #     from matplotlib import pyplot as plt
    #     from matplotlib.colors import LogNorm
    #
    #     import scopesim as sim
    #
    #     sim.rc.__config__["!SIM.file.local_packages_path"] = "F:/Work/irdb"
    #
    #     cmds = sim.UserCommands(use_instrument="MICADO",
    #                             set_modes=["SCAO", "SPEC_3000x20"])
    #     cmds["!OBS.dit"] = 3600  # [s] Exposure time (more or less)
    #     cmds["!OBS.filter_name_fw1"] = "Spec_HK"  # Spec_IJ, Spec_HK
    #     cmds["!OBS.filter_name_fw2"] = "open"
    #     cmds["!SIM.spectral.spectral_bin_width"] = 1e-4
    #
    #     micado = sim.OpticalTrain(cmds)
    #
    #     # Turn off the effects that are not needed (see below for more on this)
    #     micado["skycalc_atmosphere"].include = False
    #     micado["telescope_reflection"].include = False
    #
    #     # Set to False to test if everything is working, then flip to True
    #     USE_FULL_DETECTOR = False
    #     micado["full_detector_array"].include = USE_FULL_DETECTOR
    #     micado["detector_window"].include = not USE_FULL_DETECTOR
    #
    #     src_lines = mic_sc.line_list(unit_flux=3e23, dwave=1e-1, smoothing_fwhm=None)
    #
    #     micado.observe(src_lines)
    #
    #     plt.figure(figsize=(15, 15))
    #     plt.imshow(micado.image_planes[0].data, norm=LogNorm())
    #     plt.colorbar()
    #     plt.show()
    #
    #     fl = micado.fov_manager._fovs_list
    #     for fov in fl[1:3]:
    #         w = fov.spectra[0].model.points[0]
    #         f = fov.spectra[0].model.lookup_table
    #         plt.plot(w, f)
    #     plt.show()
    #
    #     plt.figure(figsize=(15, 15))
    #     plt.imshow(fl[3].cube.data.sum(axis=1), norm=LogNorm())
    #     plt.show()



"""Flat lamp for MICADO."""

import numpy as np
from astropy.io import fits
import astropy.units as u

from ..rc import Source


def flatlamp(
        width=16.384,
        height=16.384,
        pixel_scale=1.024,
        amplitude=1000.0,
        fraction=0.9,
    ):
    """
    A flatlamp.

    Default parameters selected such that:
    - image is 16x16
    - covering a single 4k by 4k detector.

    Parameters
    ----------
    width : float
        [arcsec] width of the 'screen'
    height : float
        [arcsec] width of the 'screen'
    pixel_scale : float
        [arcsec/pix] pixel scale
    amplitude : float
        [None] arbitrary scale factor
    fraction : float
        [None] fraction of flux at the edge

    Returns
    -------
    src: scopesim.Source
    """

    # 16x16 by default
    sx, sy = int(width / pixel_scale), int(height / pixel_scale)
    ax = np.arange(sx)
    ay = np.arange(sy)
    xx, yy = np.meshgrid(ax, ay, sparse=True)

    crpx = (width - 1) / 2
    crpy = (height - 1) / 2

    # * 1. is necessary for the normalization later
    image = - ((xx - crpx) ** 2 + (yy - crpy) ** 2) * 1.

    imin = image.min()
    imax = image.max()
    image = (1. - fraction) * (image - imin) / (imax - imin) + fraction
    image *= amplitude

    hdu = fits.ImageHDU(data=image)

    hdu.header["CRPIX1"] = crpx + 1
    hdu.header["CRPIX2"] = crpy + 1
    hdu.header["CRVAL1"] = 0
    hdu.header["CRVAL2"] = 0
    hdu.header["CDELT1"] = pixel_scale / 3600
    hdu.header["CDELT2"] = pixel_scale / 3600
    hdu.header["CUNIT1"] = "deg"
    hdu.header["CUNIT2"] = "deg"
    hdu.header["CTYPE1"] = "RA---TAN"
    hdu.header["CTYPE2"] = "DEC--TAN"

    # number_of_points should be at least 200 or so
    number_of_points = 230
    sm = 1.2419516582773063
    spectra = sm * np.ones(number_of_points)

    sa = 1.0200000e+02
    se = 3.3502039e+05
    lam = np.logspace(np.log(sa), np.log(se), base=np.e,
                      num=number_of_points) * u.Angstrom

    return Source(
            image_hdu=hdu,
            spectra=spectra,
            lam=lam,
        )



# -*- coding: utf-8 -*-
"""Tests for stellar.clusters"""

import pytest
from pytest import approx
from astropy import units as u
from astropy.table import Table

from spextra import Spextrum

from scopesim_templates.rc import Source
from scopesim_templates.stellar import cluster


@pytest.fixture(scope="module")
def basic_cluster():
    return cluster()


class TestCluster:
    @pytest.mark.usefixtures("basic_cluster")
    def test_it_works(self, basic_cluster):
        assert isinstance(basic_cluster, Source)

    @pytest.mark.usefixtures("basic_cluster")
    def test_spectra_are_correct_type(self, basic_cluster):
        if (hasattr(basic_cluster.fields[0], "field") and
                not isinstance(basic_cluster.fields[0], Table)):
            assert all(isinstance(spec, Spextrum)
                       for spec in basic_cluster.spectra.values())
        else:
            assert all(isinstance(spec, Spextrum)
                       for spec in basic_cluster.spectra)

    @pytest.mark.usefixtures("basic_cluster")
    def test_spectra_waverange(self, basic_cluster):
        # TODO: don't know if it makes sense to test like this
        if (hasattr(basic_cluster.fields[0], "field") and
                not isinstance(basic_cluster.fields[0], Table)):
            wvrng_min = [s.waverange.min().to(u.um).value
                         for s in basic_cluster.spectra.values()]
            wvrng_max = [s.waverange.max().to(u.um).value
                         for s in basic_cluster.spectra.values()]
        else:
            wvrng_min = [s.waverange.min().to(u.um).value
                         for s in basic_cluster.spectra]
            wvrng_max = [s.waverange.max().to(u.um).value
                         for s in basic_cluster.spectra]
        assert all(wave == approx(0.115) for wave in wvrng_min)
        assert all(wave == approx(2.5) for wave in wvrng_max)

    @pytest.mark.usefixtures("basic_cluster")
    def test_field_is_correct_type(self, basic_cluster):
        assert (isinstance(getattr(basic_cluster.fields[0], "field",
                                   basic_cluster.fields[0]), Table)
                or isinstance(basic_cluster.fields[0], Table))
        assert len(basic_cluster.fields[0]) > 0

    @pytest.mark.usefixtures("basic_cluster")
    def test_as_many_spectra_as_unique_sptypes(self, basic_cluster):
        uniques = set(basic_cluster.fields[0]["spec_types"])
        assert len(uniques) == len(basic_cluster.spectra)



import numpy as np
from astropy import units as u
from synphot import BlackBody1D, SourceSpectrum, Empirical1D

from ..rc import Source


def pinhole_mask(x, y, waves, temperature=1500*u.K, sum_factor=1):
    """
    Creates a Source object with point sources at coordinates (x, y) in the slit

    Parameters
    ----------
    x, y : array-like
        [arcsec] Coords relative to centre of FOV (i.e. optical axis)
    waves : array-like, u.Quantity
        [um] Array of wavelength values.
    temperature : float, u.Quantity, optional
        [deg K] Temperature of WCU blackbody spectrum.
    sum_factor : float, optional
        Rescale the spectrum to sum to this value


    Returns
    -------
    src : scopesim.Source


    Examples
    --------
    Pin-holes every 0.5 arcsec along the MICADO long-slit
    ::
        x = np.arange(-1.5, 13.51, 0.5)
        y = np.zeros_like(x)
        waves = np.arange(0.7, 2.5, 0.001) * u.um

        src = pinhole_mask(x, y, waves, sum_factor=9001)

    A grid of pin-holes every 5 arcsec for the MICADO 4mas IMG mode
    ::
        dr = np.arange(-25, 26, 5)      # [arcsec]
        x, y = np.meshgrid(dr, dr)
        x, y = x.flatten(), y.flatten()
        waves = np.arange(0.7, 2.5, 0.001) * u.um

        src = pinhole_mask(x, y, waves, sum_factor=9001)

    """
    if not isinstance(waves, u.Quantity):
        waves *= u.um
    if not isinstance(temperature, u.Quantity):
        temperature *= u.K

    blackbody_spectrum = BlackBody1D(temperature=temperature)
    flux = blackbody_spectrum(waves.to(u.AA))
    flux *= sum_factor / np.trapz(flux)
    spec = SourceSpectrum(Empirical1D, points=waves, lookup_table=flux)
    src = Source(x=x, y=y, ref=np.ones_like(x), weight=np.ones_like(x),
                 spectra=[spec])

    return src



from os.path import exists
import pytest
import synphot
from pytest import raises
from astropy import units as u
from astropy.io import fits
import numpy as np
from matplotlib.colors import LogNorm
from scopesim.utils import figure_factory

from scopesim_templates.misc import misc
from scopesim_templates.tests.pyobjects import source_objects as so
from scopesim_templates.tests.pyobjects.source_objects import starting_star_imagehdu
from scopesim_templates.rc import Source, load_example_optical_train

METIS_FILTER_PATH = r"F:/Work/irdb/METIS/filters/TC_filter_H2O-ice.dat"

PLOTS = False


class TestSourceFromImageHDU:
    def test_initialises_for_basic_imagehdu(self):
        hdu = so._basic_imagehdu()
        src = misc.source_from_imagehdu(image_hdu=hdu, filter_name="J",
                                        pixel_unit_amplitude=20*u.mag)

        assert isinstance(src, Source)

    def test_raises_error_with_no_units_provided(self):
        hdu = so._basic_imagehdu()
        with raises(ValueError):
            src = misc.source_from_imagehdu(image_hdu=hdu, filter_name="J",
                                            pixel_unit_amplitude=20)

    def test_is_happy_with_only_bunit(self):
        hdu = so._basic_imagehdu()
        hdu.header["BUNIT"] = "erg s-1 cm-2 um-1"
        src = misc.source_from_imagehdu(image_hdu=hdu, filter_name="J")

        assert isinstance(src, Source)

    def test_is_happy_with_spanish_vo_filter_name(self):
        hdu = so._basic_imagehdu()
        filter_name = "Generic/Johnson_UBVRIJHKL.N"
        src = misc.source_from_imagehdu(image_hdu=hdu, filter_name=filter_name,
                                        pixel_unit_amplitude=20 * u.Jy)

        assert isinstance(src, Source)

    @pytest.mark.skipif(not exists(METIS_FILTER_PATH),
                        reason="Test only works on Kieran's local machine")
    def test_is_happy_with_metis_filter_and_bunit(self):
        hdu = so._basic_imagehdu()
        hdu.header["BUNIT"] = "Jy"
        src = misc.source_from_imagehdu(image_hdu=hdu,
                                        filter_name=METIS_FILTER_PATH)

        assert isinstance(src, Source)

    def test_actually_produces_image(self, starting_star_imagehdu):
        src1 = misc.source_from_imagehdu(
            starting_star_imagehdu,
            filter_name="J",
            pixel_unit_amplitude=10e12 * u.Jy)

        width_height = 4096
        opt = load_example_optical_train()
        opt['psf'].include = False
        opt.cmds["!OBS.psf_fwhm"] = 0.01
        opt.cmds["!TEL.area"] = 1000 * u.m**2
        opt.cmds["!INST.pixel_scale"] = 0.004
        opt.cmds["!INST.plate_scale"] = 0.4
        opt.cmds["!DET.width"] = width_height
        opt.cmds["!DET.height"] = width_height
        opt.cmds["!DET.dit"] = 30
        opt.cmds["!DET.ndit"] = 120

        opt["source_fits_keywords"].include = False

        opt.observe(src1)
        hdul = opt.readout()[0]

        data = hdul[1].data
        # Is the background okay?
        assert 500 < np.median(data) < 5000
        # Is there a bright source?
        assert data.max() > 10000000
        # Is the bright source approximately in the center.
        x_cen, y_cen = np.unravel_index(data.argmax(), data.shape)
        # TODO: Figure out why source is not in the center!
        fudge = 100
        assert width_height / 2 - fudge < x_cen < width_height / 2 + fudge
        assert width_height / 2 - fudge < y_cen < width_height / 2 + fudge

        if PLOTS:
            fig, ax = figure_factory(3, 1)
            ax[0].imshow(src1.fields[0].data)
            ax[1].imshow(opt.image_planes[0].data, norm=LogNorm())
            ax[2].imshow(hdul[1].data, norm=LogNorm())
            fig.show()

    def test_scales(self, starting_star_imagehdu):
        """Test whether source_from_imagehdu scales w/ pixel_unit_amplitude."""
        flux1 = 10e12 * u.Jy
        src1 = misc.source_from_imagehdu(
            starting_star_imagehdu,
            filter_name="J",
            pixel_unit_amplitude=flux1)

        src2 = misc.source_from_imagehdu(
            starting_star_imagehdu,
            filter_name="J",
            pixel_unit_amplitude=flux1 * 10)

        width_height = 4096
        opt = load_example_optical_train()
        opt['psf'].include = False
        opt.cmds["!OBS.psf_fwhm"] = 0.01
        opt.cmds["!TEL.area"] = 1000 * u.m**2
        opt.cmds["!INST.pixel_scale"] = 0.004
        opt.cmds["!INST.plate_scale"] = 0.4
        opt.cmds["!DET.width"] = width_height
        opt.cmds["!DET.height"] = width_height
        opt.cmds["!DET.dit"] = 30
        opt.cmds["!DET.ndit"] = 120

        opt["source_fits_keywords"].include = False

        opt.observe(src1)
        hdul1 = opt.readout()[0]
        data1 = hdul1[1].data

        opt.observe(src2, update=True)
        hdul2 = opt.readout()[0]
        data2 = hdul2[1].data

        if PLOTS:
            fig, ax = figure_factory(1, 1)
            ax.imshow(data2 / data1)
            fig.show()

        max1 = data1.max()
        max2 = data2.max()
        assert 9 < max2 / max1 < 11


class TestPointSource:
    def test_initialize_from_spextra(self):
        src = misc.point_source("pickles/a0v", amplitude=16)

        assert isinstance(src, Source)

    def test_initialize_from_synphot(self):
        sp = synphot.SourceSpectrum(synphot.Empirical1D, points=[1000] * 4, lookup_table=[1] * 4)
        src = misc.point_source(sed=sp)
        assert isinstance(src, Source)


class TestUniformSource:
    def test_initialize_from_spextra(self):
        src = misc.uniform_source("pickles/a0v", amplitude=16)

        assert isinstance(src, Source)

    def test_initialize_from_synphot(self):
        sp = synphot.SourceSpectrum(synphot.Empirical1D, points=[1000] * 4, lookup_table=[1] * 4)
        src = misc.uniform_source(sed=sp)
        assert isinstance(src, Source)


def test_poorman_cube_source_is_working():
    cube = so._make_dummy_cube(scale=0.2, wave_unit=u.AA, ref_wave=5000,
                            wave_step=1, wave_type="WAVE", bunit="erg / (s cm2 Angstrom)")

    hdul = fits.HDUList(cube)
    cube_source = misc.poorman_cube_source(hdu=hdul, ext=0)

    assert isinstance(cube_source, Source)




import pytest

import numpy as np
from astropy import units as u
from matplotlib import pyplot as plt
from matplotlib.colors import LogNorm

from scopesim_templates.micado import viking_fields as vf
from scopesim_templates.rc import load_example_optical_train

PLOTS = False


class TestVikingCatalogues:
    @pytest.mark.parametrize("cat_id", [("1"), ("2"), ("3")])
    def test_load_galaxies_source(self, cat_id):
        gal_src = vf.load_galaxies_source(cat_id)

        if PLOTS:
            plt.figure(figsize=(20, 20))
            n = 10
            for i, r in enumerate(np.random.randint(0, 1131, n*n)):
                plt.subplot(n, n, i+1)
                plt.imshow(gal_src.fields[r].data, norm=LogNorm())
            plt.show()

        assert len(gal_src.fields) > 1
        assert len(gal_src.spectra) == 1
        assert len(gal_src.fields) == len(gal_src._meta_dicts)

    @pytest.mark.parametrize("cat_id",
                             [("stdstar"),
                              ("illum"),
                              ("science")])
    def test_load_stars_source(self, cat_id):
        star_src = vf.load_stars_source(cat_id)

        if PLOTS:
            plt.figure(figsize=(10, 10))
            x, y = star_src.fields[0]["x"], star_src.fields[0]["y"]
            size = 2.5*np.log10(star_src.fields[0]["weight"])
            size -= min(size)
            plt.scatter(x, y, s=size)
            plt.show()

        assert len(star_src.fields) == 1
        assert len(star_src.fields[0]) > 1
        assert len(star_src.spectra) == 1
        assert len(star_src.fields) == len(star_src._meta_dicts)

    def test_load_viking_field(self):
        viking_src = vf.viking_field()

        gals = viking_src.image_fields
        gal_x = [gal.header["CRVAL1"] * 3600 for gal in gals]
        gal_y = [gal.header["CRVAL2"] * 3600 for gal in gals]

        stars = viking_src.table_fields
        star_x, star_y = stars[0]["x"].data, stars[0]["y"].data

        if PLOTS:
            plt.figure(figsize=(10, 10))
            plt.scatter(gal_x, gal_y, c="b")
            plt.scatter(star_x, star_y, c="r")
            plt.show()

        assert len(viking_src.fields) > 1
        assert len(viking_src.fields) == len(gals) + len(stars)


class TestWithScopeSim:
    def test_runs_through_scopesim_basic_instrument(self):
        # configure basic_instrument to MICADO specs
        opt = load_example_optical_train()
        opt.cmds["!OBS.psf_fwhm"] = 0.01
        opt.cmds["!TEL.area"] = 1000 * u.m**2
        opt.cmds["!INST.pixel_scale"] = 0.004
        opt.cmds["!INST.plate_scale"] = 0.4
        opt.cmds["!DET.width"] = 4096
        opt.cmds["!DET.height"] = 4096
        opt.cmds["!DET.dit"] = 30
        opt.cmds["!DET.ndit"] = 120

        opt["source_fits_keywords"].include = False

        src = vf.viking_field()
        opt.observe(src)
        hdul = opt.readout()[0]

        imp_im = opt.image_planes[0].data
        imp_im += np.min(imp_im)

        if PLOTS:
            plt.figure(figsize=(15, 8))
            plt.subplot(121)
            plt.imshow(opt.image_planes[0].data, norm=LogNorm())
            plt.subplot(122)
            plt.imshow(hdul[1].data, norm=LogNorm())
            plt.show()

        assert isinstance(imp_im, np.ndarray)
        assert np.max(imp_im) > 1



import numpy as np
from collections.abc import Iterable


def nearest_spec_type(spec, cat_tbl):
    """
    Find the closest available spectral type in a catalogue.

    Parameters
    ----------
    spec : str
        Stellar spectral type
    cat_tbl
        Pickles catalogue table: ``pyckles.SpectralLibrary("pickles").table``

    Returns
    -------
    new_spec : str
        Closest spectral type in the Pickles catalogue

    """
    if isinstance(spec, Iterable) and not isinstance(spec, str):
        return [nearest_spec_type(spt, cat_tbl) for spt in spec]

    if spec in cat_tbl["name"]:
        return spec

    if spec[:2] in cat_tbl["luminosity"]:
        mask = np.where(cat_tbl["luminosity"] == spec[:2])[0]
        avail_evols = cat_tbl["evolution"][mask]
        roman = spec[4:] if spec[2] == "." else spec[2:]
        spec_evol = roman_to_arabic(roman)
        new_evol = avail_evols[np.argmin(np.abs(avail_evols - spec_evol))]
        new_spec = spec[:2] + arabic_to_roman(new_evol).upper()
        return new_spec

    if spec[0] in "OBAFGKM":
        cat_lums = cat_tbl["luminosity"]
        mask = np.where([spec[0] == lum[0] for lum in cat_lums])[0]
        avail_lums = cat_tbl["luminosity"][mask]
        avail_class = np.array([spt[1] for spt in avail_lums]).astype(int)
        avail_class = np.unique(avail_class)
        spec_class = int(spec[1])
        new_class = avail_class[np.argmin(np.abs(avail_class - spec_class))]
        new_spec = spec[0] + str(new_class) + spec[2:]
        return nearest_spec_type(new_spec, cat_tbl)

    raise ValueError(f"{spec} was not even close to a Pickles entry")


def roman_to_arabic(num: str) -> int:
    """Convert Roman numerals (I-VI) to integer."""
    dic = {"i": 1, "ii": 2, "iii": 3, "iv": 4, "v": 5, "vi": 6}
    return dic[num.lower()]


def arabic_to_roman(num: int) -> str:
    """Convert integer to Roman numerals (I-VI)."""
    dic = {1: "i", 2: "ii", 3: "iii", 4: "iv", 5: "v", 6: "vi"}
    return dic[num]



from .darkness import darkness as _darkness
from .flatlamp import flatlamp as _flatlamp
from .pinhole_masks import pinhole_mask as _pinhole_mask
from .cluster import cluster as _cluster
from .viking_fields import viking_fields
from ..utils.general_utils import add_function_call_str

from . import spectral_calibrations

@add_function_call_str
def darkness():
    return _darkness()


@add_function_call_str
def empty_sky():
    # Include empty_sky for backwards compatibility reasons.
    # TODO: Remove empty_sky when it is not used in MicadoWISE anymore.
    return _darkness()


@add_function_call_str
def flatlamp():
    return _flatlamp()


@add_function_call_str
def pinhole_mask():
    return _pinhole_mask()


@add_function_call_str
def cluster():
    return _cluster()


__all__ = [darkness, flatlamp, empty_sky, cluster, pinhole_mask]



import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm
import pytest

from astropy import units as u
from synphot import SourceSpectrum

from scopesim_templates.rc import Source
from scopesim_templates.extragalactic import galaxies

PLOTS = False


class TestGalaxy:
    def test_redshift(self):
        sp0 = galaxies.galaxy("kc96/s0", z=0, amplitude=10*u.ABmag,
                              filter_curve="g", pixel_scale=0.05, r_eff=2.5,
                              n=4, ellip=0.5, theta=45, extend=3).spectra[0]
        sp1 = galaxies.galaxy("kc96/s0", z=1, amplitude=10*u.ABmag,
                              filter_curve="g", pixel_scale=0.05, r_eff=2.5,
                              n=4, ellip=0.5, theta=45, extend=3).spectra[0]
        assert sp0.waveset.max() < sp1.waveset.max()
        kwargs = {"filter_curve": "g", "system_name": "AB"}
        assert sp0.get_magnitude(**kwargs) == sp1.get_magnitude(**kwargs)

    @pytest.mark.parametrize("amp", [10*u.ABmag, 15*u.ABmag, 20*u.ABmag])
    def test_scaling(self, amp):
        gal = galaxies.galaxy("kc96/s0", z=0, amplitude=amp,
                              filter_curve="g", pixel_scale=0.05, 
                              r_eff=2.5, n=4, ellip=0.5, theta=45, extend=3)
        kwargs = {"filter_curve": "g", "system_name": "AB"}
        assert gal.spectra[0].get_magnitude(**kwargs) == amp


class TestElliptical:
    def test_it_works(self):
        gal = galaxies.elliptical(125, 1, "V", 10 * u.ABmag, n=1,
                                  spectrum="NGC_0584")
        assert isinstance(gal, Source)

        if PLOTS:
            plt.subplot(121)
            wave = np.arange(0.3, 2.5, 0.001) * u.um
            plt.loglog(wave, gal.spectra[0](wave))

            plt.subplot(122)
            plt.imshow(gal.fields[0].data, norm=LogNorm())
            plt.colorbar()
            plt.show()


class TestSpiralTwoComponent:
    def test_returns_source_object(self):
        src = galaxies.spiral_two_component()
        assert isinstance(src, Source)
        assert len(src.fields) == 2
        assert src.fields[0].data.shape == (1200, 1200)
        assert isinstance(src.spectra[1], SourceSpectrum)

    def test_rescale_flux(self):
        src = galaxies.spiral_two_component()
